import random
import re
import unittest

from datetime import datetime
from itertools import cycle
from typing import Pattern, Tuple
from unittest import mock

import freezegun
import pytest

from validators.i18n.es import es_cif as is_cif
from validators.i18n.es import es_nie as is_nie
from validators.i18n.es import es_nif as is_nif

from faker import Factory, Faker
from faker.providers.ssn.el_GR import tin_checksum as gr_tin_checksum
from faker.providers.ssn.en_CA import checksum as ca_checksum
from faker.providers.ssn.es_CL import rut_check_digit as cl_rut_checksum
from faker.providers.ssn.es_CO import nit_check_digit
from faker.providers.ssn.es_MX import curp_checksum as mx_curp_checksum
from faker.providers.ssn.es_MX import ssn_checksum as mx_ssn_checksum
from faker.providers.ssn.et_EE import checksum as et_checksum
from faker.providers.ssn.fi_FI import Provider as fi_Provider
from faker.providers.ssn.fr_FR import calculate_checksum as fr_calculate_checksum
from faker.providers.ssn.hr_HR import checksum as hr_checksum
from faker.providers.ssn.it_IT import checksum as it_checksum
from faker.providers.ssn.lv_LV import Provider as lv_Provider
from faker.providers.ssn.no_NO import Provider as no_Provider
from faker.providers.ssn.no_NO import checksum as no_checksum
from faker.providers.ssn.pl_PL import calculate_month as pl_calculate_mouth
from faker.providers.ssn.pl_PL import checksum as pl_checksum
from faker.providers.ssn.pt_BR import checksum as pt_checksum
from faker.providers.ssn.ro_RO import ssn_checksum as ro_ssn_checksum
from faker.providers.ssn.ro_RO import vat_checksum as ro_vat_checksum
from faker.providers.ssn.uk_UA import Provider as uk_Provider
from faker.providers.ssn.zh_TW import checksum as tw_checksum
from faker.utils.checksums import luhn_checksum


class TestSvSE(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("sv_SE")
        Faker.seed(0)

    def partial_sum(self, number, mult_factor):
        quotient, remainder = divmod(number * mult_factor, 10)
        return quotient + remainder

    def ssn_checksum(self, ssn):
        """Validates the checksum digit and returns a Boolean"""
        ssn = ssn.replace("-", "")
        if len(ssn) == 12:
            ssn = ssn[2:]
        if len(ssn) != 10:
            return False

        mult_factors = cycle([2, 1])
        final_sum = sum(self.partial_sum(int(char), mf) for char, mf in zip(ssn[:9], mult_factors))
        chksum = -final_sum % 10
        return chksum == int(ssn[-1])

    def validate_date_string(self, date_str):
        date_len = len(date_str)
        if date_len == 6:
            year_fmt = "%y"
        elif date_len == 8:
            year_fmt = "%Y"
        else:
            return False

        try:
            if date_str != datetime.strptime(date_str, f"{year_fmt}%m%d").strftime(f"{year_fmt}%m%d"):
                raise ValueError
            return True
        except ValueError:
            return False

    def test_pers_id_short_with_dash(self):
        """Regression case that ensures previous implementations work as-is"""
        for _ in range(100):
            pers_id = self.fake.ssn()
            assert re.search(r"\d{6}-\d{4}", pers_id)
            assert self.validate_date_string(pers_id[:6]) is True
            assert self.ssn_checksum(pers_id) is True

    def test_pers_id_short_no_dash(self):
        for _ in range(100):
            pers_id = self.fake.ssn(dash=False)
            assert re.search(r"\d{10}", pers_id)
            assert self.validate_date_string(pers_id[:6]) is True
            assert self.ssn_checksum(pers_id) is True

    def test_pers_id_long_with_dash(self):
        for _ in range(100):
            pers_id = self.fake.ssn(long=True)
            assert re.search(r"\d{8}-\d{4}", pers_id)
            assert self.validate_date_string(pers_id[:8]) is True
            assert self.ssn_checksum(pers_id) is True

    def test_pers_id_long_no_dash(self):
        for _ in range(100):
            pers_id = self.fake.ssn(long=True, dash=False)
            assert re.search(r"\d{12}", pers_id)
            assert self.validate_date_string(pers_id[:8]) is True
            assert self.ssn_checksum(pers_id) is True

    def test_org_id_short_with_dash(self):
        for _ in range(100):
            org_id = self.fake.org_id()
            assert re.search(r"\d{6}-\d{4}", org_id)
            assert int(org_id[2:4]) >= 20
            assert self.ssn_checksum(org_id) is True

    def test_org_id_short_no_dash(self):
        for _ in range(100):
            org_id = self.fake.org_id(dash=False)
            assert re.search(r"\d{10}", org_id)
            assert int(org_id[2:4]) >= 20
            assert self.ssn_checksum(org_id) is True

    def test_org_id_long_with_dash(self):
        for _ in range(100):
            org_id = self.fake.org_id(long=True)
            assert re.search(r"\d{8}-\d{4}", org_id)
            assert int(org_id[4:6]) >= 20
            assert self.ssn_checksum(org_id) is True

    def test_org_id_long_no_dash(self):
        for _ in range(100):
            org_id = self.fake.org_id(long=True, dash=False)
            assert re.search(r"\d{12}", org_id)
            assert int(org_id[4:6]) >= 20
            assert self.ssn_checksum(org_id) is True

    def test_vat_id(self):
        for _ in range(100):
            vat_id = self.fake.vat_id()
            assert re.search(r"SE\d{12}", vat_id)
            assert int(vat_id[2]) in (1, 2, 3, 5, 6, 7, 8, 9)
            assert int(vat_id[6:8]) >= 20

    def test_org_and_vat_id(self):
        for _ in range(100):
            oid, vid = self.fake.org_and_vat_id()
            assert oid.replace("-", "")[-10:] == vid[4:-2]
            assert re.search(r"SE\d{12}", vid)


class TestBgBG(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("bg_BG")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^BG\d{9,10}$", self.fake.vat_id())


class TestCsCZ(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("cs_CZ")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^CZ\d{8,10}$", self.fake.vat_id())

    def test_birth_number(self):
        for _ in range(100):
            birth_number = self.fake.birth_number()
            assert len(birth_number) in [10, 11]
            assert birth_number[6] == "/"
            assert int(birth_number.replace("/", "")) % 11 == 0


class TestSkSK(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("sk_SK")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^SK\d{10}$", self.fake.vat_id())

    def test_birth_number(self):
        for _ in range(100):
            birth_number = self.fake.birth_number()
            assert len(birth_number) in [10, 11]
            assert birth_number[6] == "/"
            assert int(birth_number.replace("/", "")) % 11 == 0


class TestDeAT(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("de_AT")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^ATU\d{8}$", self.fake.vat_id())


class TestElCY(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("el_CY")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^CY\d{9}\w$", self.fake.vat_id())


class TestElGr(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("el_GR")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            prefix = random.choice([True, False])
            vat_id = self.fake.vat_id(prefix=prefix)
            assert re.search(r"^(EL)?\d{9}$", vat_id)
            assert vat_id[2 if prefix else 0] in ("7", "8", "9", "0")
            assert str(gr_tin_checksum(vat_id[2:-1] if prefix else vat_id[:-1])) == vat_id[-1]

    def test_tin(self):
        for _ in range(100):
            tin = self.fake.tin()
            assert re.search(r"^\d{9}$", tin)
            assert tin[0] in ("1", "2", "3", "4")
            assert str(gr_tin_checksum(tin[:-1])) == tin[-1]

    def test_ssn(self):
        for _ in range(100):
            ssn = self.fake.ssn()
            assert re.search(r"^\d{11}$", ssn)
            assert datetime.strptime(ssn[:6], "%d%m%y")
            assert luhn_checksum(ssn) == 0

    def test_police_id(self):
        for _ in range(100):
            assert re.search(r"^[ΑΒΕΖΗΙΚΜΝΟΡΤΥΧ]{1,2}\d{6}$", self.fake.police_id())


class TestEnCA(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("en_CA")
        Faker.seed(0)

    def test_ssn(self):
        for _ in range(100):
            sin = self.fake.ssn()

            # Ensure that generated SINs are 11 characters long
            # including spaces, consist of spaces and digits only, and
            # satisfy the validation algorithm.
            assert len(sin) == 11
            assert sin.replace(" ", "").isdigit()
            assert ca_checksum(sin) == int(sin[-1])


class TestEnUS(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("en_US")
        Faker.seed(0)

    def test_ssn(self):
        for _ in range(100):
            ssn = self.fake.ssn(taxpayer_identification_number_type="SSN")

            # Ensure that generated SINs are 11 characters long
            # including dashes, consist of dashes and digits only, and
            # satisfy these requirements:
            #
            # An United States Social Security Number
            # (SSN) is a tax processing number issued by the Internal
            # Revenue Service with the format "AAA-GG-SSSS".  The
            # number is divided into three parts: the first three
            # digits, known as the area number because they were
            # formerly assigned by geographical region; the middle two
            # digits, known as the group number; and the final four
            # digits, known as the serial number. SSNs with the
            # following characteristics are not allocated:
            #
            # 1) Numbers with all zeros in any digit group
            # (000-##-####, ###-00-####, ###-##-0000).
            #
            # 2) Numbers with 666 or 900-999 in the first digit group.
            #
            # https://en.wikipedia.org/wiki/Social_Security_number

            assert len(ssn) == 11
            assert ssn.replace("-", "").isdigit()

            [area, group, serial] = ssn.split("-")

            assert 1 <= int(area) <= 899 and int(area) != 666
            assert 1 <= int(group) <= 99
            assert 1 <= int(serial) <= 9999
            assert area != "666"

    def test_invalid_ssn(self):
        # Magic Numbers below generate '666-92-7944', '000-54-2963', '956-GG-9478', '436-00-1386',
        # and 134-76-0000 respectively. The "group" (GG) returned for '956-GG-9478 will be a random
        # number, and that random number is not in the "itin_group_numbers" List. The random GG occurs
        # even when using the same seed_instance() due to using random.choice() for GG to avoid valid
        # ITINs being returned as an invalid SSN:
        #
        # Ensure that generated SSNs are 11 characters long
        # including dashes, consist of dashes and digits only, and the tested number
        # violates the requirements below, ensuring an INVALID SSN is returned:
        #
        # A United States Social Security Number
        # (SSN) is a tax processing number issued by the Internal
        # Revenue Service with the format "AAA-GG-SSSS".  The
        # number is divided into three parts: the first three
        # digits, known as the area number because they were
        # formerly assigned by geographical region; the middle two
        # digits, known as the group number; and the final four
        # digits, known as the serial number. SSNs with the
        # following characteristics are not allocated:
        #
        # 1) Numbers with all zeros in any digit group
        # (000-##-####, ###-00-####, ###-##-0000).
        #
        # 2) Numbers with 666 or 900-999 in the first digit group.
        #
        # https://en.wikipedia.org/wiki/Social_Security_number
        #
        # ITIN explained:
        # https://www.irs.gov/individuals/international-taxpayers/general-itin-information

        itin_group_numbers = [
            70,
            71,
            72,
            73,
            74,
            75,
            76,
            77,
            78,
            79,
            80,
            81,
            82,
            83,
            84,
            85,
            86,
            87,
            88,
            90,
            91,
            92,
            94,
            95,
            96,
            97,
            98,
            99,
        ]

        self.fake.seed_instance(2432)
        ssn = self.fake.ssn(taxpayer_identification_number_type="INVALID_SSN")

        assert len(ssn) == 11
        assert ssn.replace("-", "").isdigit()
        assert ssn.startswith("666")

        self.fake.seed_instance(1514)
        ssn = self.fake.ssn(taxpayer_identification_number_type="INVALID_SSN")

        assert ssn.startswith("000")

        self.fake.seed_instance(2)
        ssn = self.fake.ssn(taxpayer_identification_number_type="INVALID_SSN")
        [area, group, serial] = ssn.split("-")

        assert 900 <= int(area) <= 999 and int(group) not in itin_group_numbers

        self.fake.seed_instance(0)
        ssn = self.fake.ssn(taxpayer_identification_number_type="INVALID_SSN")
        [area, group, serial] = ssn.split("-")

        assert int(area) < 900 and int(group) == 0

        self.fake.seed_instance(1)
        ssn = self.fake.ssn(taxpayer_identification_number_type="INVALID_SSN")
        [area, group, serial] = ssn.split("-")

        assert int(area) < 900 and int(serial) == 0

    def test_prohibited_ssn_value(self):
        # 666 is a prohibited value. The magic number selected as a seed
        # is one that would (if not specifically checked for) return an
        # SSN with an area of '666'.

        Faker.seed(19031)
        ssn = self.fake.ssn()
        [area, group, serial] = ssn.split("-")
        assert area != "666"

    def test_itin(self):
        for _ in range(100):
            itin = self.fake.ssn(taxpayer_identification_number_type="ITIN")

            # Ensure that generated SINs are 11 characters long
            # including dashes, consist of dashes and digits only, and
            # satisfy these requirements:
            #
            # An United States Individual Taxpayer Identification Number
            # (ITIN) is a tax processing number issued by the Internal
            # Revenue Service. It is a nine-digit number that always begins
            # with the number 9 and has a range of 70-88 in the fourth and
            # fifth digit. Effective April 12, 2011, the range was extended
            # to include 900-70-0000 through 999-88-9999, 900-90-0000
            # through 999-92-9999 and 900-94-0000 through 999-99-9999.
            # https://www.irs.gov/individuals/international-taxpayers/general-itin-information

            assert len(itin) == 11
            assert itin.replace("-", "").isdigit()

            [area, group, serial] = itin.split("-")

            assert 900 <= int(area) <= 999
            assert 70 <= int(group) <= 88 or 90 <= int(group) <= 92 or 94 <= int(group) <= 99
            assert 0 <= int(serial) <= 9999

    def test_ein(self):
        ein_prefix_choices = [
            "01",
            "02",
            "03",
            "04",
            "05",
            "06",
            "10",
            "11",
            "12",
            "13",
            "14",
            "15",
            "16",
            "20",
            "21",
            "22",
            "23",
            "24",
            "25",
            "26",
            "27",
            "30",
            "31",
            "32",
            "33",
            "34",
            "35",
            "36",
            "37",
            "38",
            "39",
            "40",
            "41",
            "42",
            "43",
            "44",
            "45",
            "46",
            "47",
            "48",
            "50",
            "51",
            "52",
            "53",
            "54",
            "55",
            "56",
            "57",
            "58",
            "59",
            "60",
            "61",
            "62",
            "63",
            "64",
            "65",
            "66",
            "67",
            "68",
            "71",
            "72",
            "73",
            "74",
            "75",
            "76",
            "77",
            "80",
            "81",
            "82",
            "83",
            "84",
            "85",
            "86",
            "87",
            "88",
            "90",
            "91",
            "92",
            "93",
            "94",
            "95",
            "98",
            "99",
        ]

        for _ in range(100):
            ein = self.fake.ssn(taxpayer_identification_number_type="EIN")

            # An United States An Employer Identification Number (EIN) is
            # also known as a Federal Tax Identification Number, and is
            # used to identify a business entity. EINs follow a format of a
            # two-digit prefix followed by a hyphen and a seven-digit sequence.
            # https://www.irs.gov/businesses/small-businesses-self-employed/employer-id-numbers
            #
            # Ensure that generated EINs are 10 characters long
            # including a dash, consist of dashes and digits only, and
            # satisfy these requirements:
            #
            # There are only certain EIN Prefix values assigned:
            # https://www.irs.gov/businesses/small-businesses-self-employed/how-eins-are-assigned-and-valid-ein-prefixes

            assert len(ein) == 10
            assert ein.replace("-", "").isdigit()

            [prefix, sequence] = ein.split("-")

            assert prefix in ein_prefix_choices
            assert 0 <= int(sequence) <= 9999999

    def test_bad_tin_type(self):
        with self.assertRaises(ValueError):
            self.fake.ssn(taxpayer_identification_number_type="badValue")

    def test_wrong_tin_type_case(self):
        with self.assertRaises(ValueError):
            self.fake.ssn(taxpayer_identification_number_type="ssn")


class TestEsCO(unittest.TestCase):
    def setUp(self):
        self._NUIP_REGEX: Pattern = re.compile(r"1[012]\d{8}|[1-9]\d{6,7}")
        self._NATURAL_PERSON_NIT_REGEX: Pattern = self._NUIP_REGEX
        self._CHECK_DIGIT_REGEX: Pattern = re.compile(r"\d")
        self._LEGAL_PERSON_NIT_REGEX: Pattern = re.compile(r"[89]\d{8}")
        self.fake = Faker("es_CO")
        Faker.seed(0)

    def test_nuip(self):
        for _ in range(100):
            assert self._NUIP_REGEX.fullmatch(self.fake.nuip())
            assert self._NUIP_REGEX.fullmatch(self.fake.natural_person_nit())

    def test_natural_person_nit_with_check_digit(self):
        for _ in range(100):
            natural_person_nit, check_digit = self.fake.natural_person_nit_with_check_digit().split("-")
            assert self._NATURAL_PERSON_NIT_REGEX.fullmatch(natural_person_nit)
            assert self._CHECK_DIGIT_REGEX.fullmatch(check_digit)
            assert nit_check_digit(natural_person_nit) == check_digit

    def test_legal_person_nit(self):
        for _ in range(100):
            assert self._LEGAL_PERSON_NIT_REGEX.fullmatch(self.fake.legal_person_nit())

    def test_legal_person_nit_with_check_digit(self):
        for _ in range(100):
            legal_person_nit, check_digit = self.fake.legal_person_nit_with_check_digit().split("-")
            assert self._LEGAL_PERSON_NIT_REGEX.fullmatch(legal_person_nit)
            assert self._CHECK_DIGIT_REGEX.fullmatch(check_digit)
            assert nit_check_digit(legal_person_nit) == check_digit

    def test_nit_check_digit(self):
        # NITs and check digits of some Colombian state entities.
        # Source: <https://www.funcionpublica.gov.co/web/sigep/entidades>
        for nit, check_digit in (
            ("830040256", "0"),
            ("899999003", "1"),
            ("892301483", "2"),
            ("800194600", "3"),
            ("899999403", "4"),
            ("860042945", "5"),
            ("830114475", "6"),
            ("811000231", "7"),
            ("899999027", "8"),
            ("900639630", "9"),
        ):
            with self.subTest(nit=nit, check_digit=check_digit):
                assert nit_check_digit(nit) == check_digit


class TestEsES(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("es_ES")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^ES\w\d{8}$|^ES\d{8}\w$|^ES\w\d{7}\w$", self.fake.vat_id())

    def test_nie(self):
        for _ in range(100):
            assert is_nie(self.fake.nie())

    def test_nif(self):
        for _ in range(100):
            assert is_nif(self.fake.nif())

    def test_cif(self):
        for _ in range(100):
            assert is_cif(self.fake.cif())

    def test_doi(self):
        assert len(self.fake.doi()) == 9

    def test_nuss(self):
        for _ in range(50):
            nuss = self.fake.nuss()
            assert isinstance(nuss, str)
            assert 12 == len(nuss)
        for _ in range(50):
            nuss = self.fake.nuss(company=True)
            assert isinstance(nuss, str)
            assert 11 == len(nuss)


class TestEsCA(TestEsES):
    def setUp(self):
        self.fake = Faker("es_CA")
        Faker.seed(0)


class TestEsMX:
    def setup_method(self):
        self.fake = Faker("es_MX")
        Faker.seed(0)

    def test_ssn(self):
        for _ in range(100):
            ssn = self.fake.ssn()

            assert len(ssn) == 11
            assert ssn.isnumeric()
            assert mx_ssn_checksum(map(int, ssn[:-1])) == int(ssn[-1])

    def test_curp(self):
        for _ in range(100):
            curp = self.fake.curp()

            assert len(curp) == 18
            assert re.search(r"^[A-Z]{4}\d{6}[A-Z]{6}[0A]\d$", curp)
            assert mx_curp_checksum(curp[:-1]) == int(curp[-1])

    def test_rfc_natural(self):
        for _ in range(100):
            rfc = self.fake.rfc()

            assert len(rfc) == 13
            assert re.search(r"^[A-Z]{4}\d{6}[0-9A-Z]{3}$", rfc)

    def test_rfc_legal(self):
        for _ in range(100):
            rfc = self.fake.rfc(natural=False)

            assert len(rfc) == 12
            assert re.search(r"^[A-Z]{3}\d{6}[0-9A-Z]{3}$", rfc)

    @pytest.mark.parametrize(
        "gender,pattern",
        [
            ("M", r"^[A-Z]{6}\d{8}M\d{3}$"),
            ("H", r"^[A-Z]{6}\d{8}H\d{3}$"),
            (None, r"^[A-Z]{6}\d{8}[HM]\d{3}$"),
        ],
        ids=["woman", "man", "any"],
    )
    def test_elector_code(self, gender, pattern):
        for _ in range(100):
            elector_code = self.fake.elector_code(gender=gender)

            assert len(elector_code) == 18
            assert re.search(pattern, elector_code)

    def test_elector_code_unsupported_gender(self):
        with pytest.raises(ValueError, match="Gender must be"):
            self.fake.elector_code("Z")


class TestEsCL(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("es_CL")
        Faker.seed(0)

    def test_rut(self):
        for _ in range(100):
            rut = self.fake.rut(min=10000000)
            digits, check_digit = self._extract_digits(rut)

            assert len(rut) == 12
            assert check_digit == cl_rut_checksum(digits)

    @staticmethod
    def _extract_digits(rut) -> Tuple[int, str]:
        """Extracts the digits and check digit from a formatted RUT."""
        char_filter = re.compile(r"[^0-9]")
        check_digit = rut[-1]
        digits = char_filter.sub("", rut[:-1])

        return int(digits), check_digit


class TestEtEE(unittest.TestCase):
    """Tests SSN in the et_EE locale"""

    def setUp(self):
        self.fake = Faker("et_EE")

        Faker.seed(0)

    def test_ssn_checksum(self):
        assert et_checksum([4, 4, 1, 1, 1, 3, 0, 4, 9, 2]) == 3
        assert et_checksum([3, 6, 7, 0, 1, 1, 6, 6, 2, 7]) == 8
        assert et_checksum([4, 7, 0, 0, 4, 2, 1, 5, 0, 1]) == 2
        assert et_checksum([3, 9, 7, 0, 3, 0, 4, 3, 3, 6]) == 0

    @freezegun.freeze_time("2019-03-11")
    def test_ssn(self):
        self.fake.seed_instance(1)
        value = self.fake.ssn()
        assert re.search(r"^\d{11}$", value)
        assert not value.endswith("0")

        self.fake.seed_instance(0)
        value = self.fake.ssn()

        assert re.search(r"^\d{11}$", value)
        assert value.endswith("0")

    @freezegun.freeze_time("2002-01-01")
    def test_ssn_2000(self):
        self.fake.seed_instance(0)
        value = self.fake.ssn(min_age=0, max_age=1)
        assert re.search(r"^\d{11}$", value)
        assert value[0] in ("5", "6")

    @freezegun.freeze_time("2101-01-01")
    def test_ssn_2100(self):
        self.fake.seed_instance(0)
        value = self.fake.ssn(min_age=0, max_age=1)
        assert re.search(r"^\d{11}$", value)
        assert value[0] in ("7", "8")

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^EE\d{9}$", self.fake.vat_id())


class TestFiFI(unittest.TestCase):
    """Tests SSN in the fi_FI locale"""

    def setUp(self):
        self.fake = Faker("fi_FI")
        Faker.seed(0)
        self.provider = fi_Provider

    def test_century_code(self):
        assert self.provider._get_century_code(1900) == "-"
        assert self.provider._get_century_code(1999) == "-"
        assert self.provider._get_century_code(2000) == "A"
        assert self.provider._get_century_code(2999) == "A"
        assert self.provider._get_century_code(1800) == "+"
        assert self.provider._get_century_code(1899) == "+"
        with pytest.raises(ValueError):
            self.provider._get_century_code(1799)
        with pytest.raises(ValueError):
            self.provider._get_century_code(3000)

    def test_ssn_sanity(self):
        for age in range(100):
            self.fake.ssn(min_age=age, max_age=age + 1)

    def test_valid_ssn(self):
        ssn = self.fake.ssn(artificial=False)
        individual_number = int(ssn[7:10])
        assert individual_number <= 899

    def test_artifical_ssn(self):
        ssn = self.fake.ssn(artificial=True)
        individual_number = int(ssn[7:10])
        assert individual_number >= 900

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^FI\d{8}$", self.fake.vat_id())

    @freezegun.freeze_time("2023-10-23")
    def test_ssn_without_age_range(self):
        current_year = 2023
        age = current_year - 1995
        ssn = self.fake.ssn(min_age=age, max_age=age, artificial=True)
        assert "95-" in ssn
        age = current_year - 2013
        ssn = self.fake.ssn(min_age=age, max_age=age, artificial=True)
        assert "13A" in ssn
        age = current_year - 1898
        ssn = self.fake.ssn(min_age=age, max_age=age, artificial=True)
        assert "98+" in ssn


class TestFrFR(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("fr_FR")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^FR[\w\d]{2} \d{9}$", self.fake.vat_id())

    def test_ssn(self) -> None:
        for _ in range(100):
            assert re.search(r"^\d{15}$", self.fake.ssn())

    def test_checksum(self) -> None:
        assert fr_calculate_checksum(2570533063999) == 3


class TestFrCH:
    @pytest.mark.parametrize(
        "digits,expected",
        [
            ("22500105", "CHE225001055"),
            ("60362354", "CHE603623540"),
            ("36806684", "CHE368066842"),
        ],
        ids=[
            "checksum_remainder_11",
            "checksum_remainder_10",
            "checksum_remainder_other",
        ],
    )
    def test_checksum(self, digits, expected):
        """The checksum of the Swiss UID number is calculated correctly
        given a certain input of 8 digits."""
        fake = Faker("fr_CH")
        Faker.seed(0)

        with mock.patch(
            "faker.providers.ssn.fr_CH.Provider.numerify",
            return_value=digits,
            autospec=True,
        ):
            result = fake.vat_id()
            assert result == expected


class TestEnGB(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("en_GB")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(
                r"^GB\d{3} \d{4} \d{2}(?: \d{3})?$|^GB(?:GD|HA)\d{3}$",
                self.fake.vat_id(),
            )


class TestHrHR(unittest.TestCase):
    """Tests SSN in the hr_HR locale"""

    def setUp(self):
        self.fake = Faker("hr_HR")
        Faker.seed(0)

    def test_ssn_checksum(self):
        assert hr_checksum([0, 0, 2, 2, 8, 2, 6, 9, 2, 8]) == 9
        assert hr_checksum([5, 8, 9, 3, 6, 9, 5, 1, 2, 5]) == 1
        assert hr_checksum([5, 7, 8, 0, 2, 0, 3, 4, 2, 3]) == 7
        assert hr_checksum([4, 3, 3, 3, 1, 4, 6, 7, 6, 2]) == 2
        assert hr_checksum([0, 5, 9, 3, 7, 7, 5, 9, 1, 8]) == 7
        assert hr_checksum([7, 1, 1, 4, 9, 9, 1, 2, 4, 1]) == 6

    def test_ssn(self):
        for _ in range(100):
            assert re.search(r"^\d{11}$", self.fake.ssn())

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^HR\d{11}$", self.fake.vat_id())


class TestHuHU(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("hu_HU")
        Faker.seed(0)

    def test_ssn(self):
        for _ in range(100):
            ssn = self.fake.ssn()
            assert ssn.isdigit()
            assert len(ssn) >= 10
            assert len(ssn) <= 12

        for _ in range(100):
            dob_val = (
                f"{self.fake.random_int(0, 99):02d}"
                f"{self.fake.random_int(1, 12):02d}"
                f"{self.fake.random_int(1, 31):02d}"
            )
            dob = self.fake.random.choice([None, dob_val])
            gender = self.fake.random.choice([None, "F", "M", "z"])
            try:
                ssn = self.fake.ssn(dob=dob, gender=gender)
                assert ssn.isdigit()
                assert len(ssn) >= 10
                assert len(ssn) <= 12
            except ValueError:
                pass

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^HU\d{8}$", self.fake.vat_id())


class TestItIT(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("it_IT")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^IT\d{11}$", self.fake.vat_id())

    def test_ssn(self):
        for _ in range(100):
            assert re.search(r"^[A-Z]{6}\d{2}[ABCDEHLMPRST][0-7]\d[A-Z]\d{3}[A-Z]$", self.fake.ssn())

    def test_checksum(self) -> None:
        assert it_checksum("MDDMRA80L41H501") == "R"

    def test_ssn_with_latin_chars(self):
        generator = Factory.create("it_IT")
        generator.last_name = mock.MagicMock(return_value="Foà")
        ssn = generator.ssn()
        self.assertEqual(len(ssn), 16)
        self.assertEqual(ssn[:3], "FOA")


class TestPtBR(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("pt_BR")
        Faker.seed(0)

    def test_pt_BR_ssn_checksum(self):
        assert pt_checksum([8, 8, 2, 8, 2, 1, 6, 5, 2]) == 2
        assert pt_checksum([8, 8, 2, 8, 2, 1, 6, 5, 2, 2]) == 0

    def test_pt_BR_ssn(self):
        for _ in range(100):
            assert re.search(r"^\d{11}$", self.fake.ssn())

    def test_pt_BR_cpf(self):
        for _ in range(100):
            assert re.search(r"\d{3}\.\d{3}\.\d{3}-\d{2}", self.fake.cpf())

    def test_pt_BR_rg(self):
        for _ in range(100):
            to_test = self.fake.rg()
            if "X" in to_test:
                assert re.search(r"^\d{8}X", to_test)
            else:
                assert re.search(r"^\d{9}$", to_test)


class TestNlBE(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("nl_BE")
        Faker.seed(0)

    def test_ssn(self):
        for _ in range(1000):
            ssn = self.fake.ssn()
            assert len(ssn) == 11
            gen_seq = ssn[6:9]
            gen_chksum = ssn[9:11]
            gen_seq_as_int = int(gen_seq)
            gen_chksum_as_int = int(gen_chksum)
            # Check that the sequence nr is between 1 inclusive and 998 inclusive
            assert gen_seq_as_int > 0
            assert gen_seq_as_int <= 998

            # validate checksum calculation
            # Since the century is not part of ssn, try both below and above year 2000
            ssn_below = int(ssn[0:9])
            chksum_below = 97 - (ssn_below % 97)
            ssn_above = ssn_below + 2000000000
            chksum_above = 97 - (ssn_above % 97)
            results = [chksum_above, chksum_below]
            assert gen_chksum_as_int in results


class TestNlNL(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("nl_NL")
        Faker.seed(0)

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^NL\d{9}B\d{2}$", self.fake.vat_id())


class TestNoNO(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("no_NO")
        Faker.seed(0)

    def test_no_NO_ssn_checksum(self):
        assert no_checksum([0, 1, 0, 2, 0, 3, 9, 8, 7], no_Provider.scale1) == 6
        assert no_checksum([0, 1, 0, 2, 0, 3, 9, 8, 7, 6], no_Provider.scale2) == 7

    def test_no_NO_ssn(self):
        for _ in range(100):
            ssn = self.fake.ssn()
            assert ssn.isdigit()
            assert len(ssn) == 11

    def test_no_NO_ssn_dob_passed(self):
        test_data = [("20010203", "030201"), ("19991231", "311299")]
        for date_of_birth, expected_dob_part in test_data:
            ssn = self.fake.ssn(dob=date_of_birth)
            assert ssn[:6] == expected_dob_part

    def test_no_NO_ssn_invalid_dob_passed(self):
        with pytest.raises(ValueError):
            self.fake.ssn(dob="010401")
        with pytest.raises(ValueError):
            self.fake.ssn(dob="hello_world")
        with pytest.raises(ValueError):
            self.fake.ssn(dob="001301")

    def test_no_NO_ssn_gender_passed(self):
        # Females have even number at index 8
        ssn = self.fake.ssn(gender="F")
        assert int(ssn[8]) % 2 == 0
        # Males have odd number at index 8
        ssn = self.fake.ssn(gender="M")
        assert int(ssn[8]) % 2 == 1

    def test_no_NO_ssn_invalid_gender_passed(self):
        with pytest.raises(ValueError):
            self.fake.ssn(gender="A")


class TestPlPL(unittest.TestCase):
    """Tests SSN in the pl_PL locale"""

    def setUp(self):
        self.fake = Faker("pl_PL")
        Faker.seed(0)

    def test_ssn_checksum(self):
        assert pl_checksum([0, 5, 2, 6, 2, 8, 1, 2, 3, 6]) == 5
        assert pl_checksum([8, 5, 0, 5, 0, 8, 1, 5, 5, 8]) == 7
        assert pl_checksum([4, 5, 1, 1, 1, 0, 0, 2, 4, 3]) == 3
        assert pl_checksum([9, 1, 0, 7, 2, 6, 1, 4, 8, 7]) == 3
        assert pl_checksum([8, 1, 1, 2, 1, 4, 1, 1, 8, 7]) == 6

    def test_calculate_month(self):
        assert pl_calculate_mouth(datetime.strptime("1 1 1900", "%m %d %Y")) == 1
        assert pl_calculate_mouth(datetime.strptime("12 1 1900", "%m %d %Y")) == 12
        assert pl_calculate_mouth(datetime.strptime("1 1 1999", "%m %d %Y")) == 1

        assert pl_calculate_mouth(datetime.strptime("1 1 2000", "%m %d %Y")) == 21
        assert pl_calculate_mouth(datetime.strptime("12 1 2000", "%m %d %Y")) == 32
        assert pl_calculate_mouth(datetime.strptime("1 1 2099", "%m %d %Y")) == 21

        assert pl_calculate_mouth(datetime.strptime("1 1 2100", "%m %d %Y")) == 41
        assert pl_calculate_mouth(datetime.strptime("12 1 2100", "%m %d %Y")) == 52
        assert pl_calculate_mouth(datetime.strptime("1 1 2199", "%m %d %Y")) == 41

        assert pl_calculate_mouth(datetime.strptime("1 1 2200", "%m %d %Y")) == 61
        assert pl_calculate_mouth(datetime.strptime("12 1 2200", "%m %d %Y")) == 72
        assert pl_calculate_mouth(datetime.strptime("1 1 2299", "%m %d %Y")) == 61

    def test_ssn(self):
        for _ in range(100):
            assert re.search(r"^\d{11}$", self.fake.ssn())

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^PL\d{10}$", self.fake.vat_id())


class TestEnPh(unittest.TestCase):
    num_sample_runs = 1000

    def setUp(self):
        self.sss_pattern: Pattern = re.compile(r"^\d{2}-\d{7}-\d$")
        self.gsis_pattern: Pattern = re.compile(r"^\d{11}$")
        self.philhealth_pattern: Pattern = re.compile(r"^\d{2}-\d{9}-\d$")
        self.pagibig_pattern: Pattern = re.compile(r"^\d{4}-\d{4}-\d{4}$")
        self.umid_pattern: Pattern = re.compile(r"^\d{4}-\d{7}-\d$")
        self.setup_faker()

    def setup_faker(self):
        self.fake = Faker("en_PH")
        Faker.seed(0)

    def test_PH_sss(self):
        for i in range(self.num_sample_runs):
            assert self.sss_pattern.match(self.fake.sss())

    def test_PH_gsis(self):
        for i in range(self.num_sample_runs):
            assert self.gsis_pattern.match(self.fake.gsis())

    def test_PH_philhealth(self):
        for i in range(self.num_sample_runs):
            assert self.philhealth_pattern.match(self.fake.philhealth())

    def test_PH_pagibig(self):
        for i in range(self.num_sample_runs):
            assert self.pagibig_pattern.match(self.fake.pagibig())

    def test_PH_umid(self):
        for i in range(self.num_sample_runs):
            assert self.umid_pattern.match(self.fake.umid())


class TestFilPh(TestEnPh):
    def setup_faker(self):
        self.fake = Faker("fil_PH")
        Faker.seed(0)


class TestThTH(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("th_TH")
        Faker.seed(0)

    def test_ssn(self):
        for _ in range(100):
            assert re.search(r"^[1-8]-[1-9]\d{3}-\d{5}-\d{2}-\d$", self.fake.ssn())

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^[1-8]-[1-9]\d{3}-\d{5}-\d{2}-\d$", self.fake.vat_id())


class TestTlPh(TestEnPh):
    def setup_faker(self):
        self.fake = Faker("tl_PH")
        Faker.seed(0)


class TestTrTr(unittest.TestCase):
    num_sample_runs = 10

    def setUp(self):
        self.fake = Faker("tr_TR")
        self.samples = [self.fake.ssn() for _ in range(self.num_sample_runs)]
        Faker.seed(0)

    def first_part_non_zero(self):
        for sample in self.samples:
            assert sample[0] != 0

    def compare_first_ten_and_last_part(self):
        for sample in self.samples:
            first_ten_number = sample[:-1]
            last_part = sample[-1]
            assert sum(int(x) for x in f"{first_ten_number}") % 10 == last_part


class TestEnIn(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("en_IN")
        Faker.seed(0)
        test_samples = 10
        self.aadhaar_ids = [self.fake.aadhaar_id() for _ in range(test_samples)]

    def test_length(self):
        for aadhaar_id in self.aadhaar_ids:
            assert len(aadhaar_id) == 12

    def test_first_digit_non_zero(self):
        for aadhar_id in self.aadhaar_ids:
            assert aadhar_id[0] != "0"

    def test_valid_luhn(self):
        for aadhaar_id in self.aadhaar_ids:
            assert luhn_checksum(aadhaar_id) == 0


class TestZhCN(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("zh_CN")
        Faker.seed(0)

    def test_zh_CN_ssn(self):
        for _ in range(100):
            ssn = self.fake.ssn()
            assert len(ssn) == 18

    def test_zh_CN_ssn_invalid_gender_passed(self):
        with pytest.raises(ValueError):
            self.fake.ssn(gender="X")
        with pytest.raises(ValueError):
            self.fake.ssn(gender="*")
        with pytest.raises(ValueError):
            self.fake.ssn(gender="22")

    def test_zh_CN_ssn_gender_passed(self):
        # Females have even number at index 17
        ssn = self.fake.ssn(gender="F")
        assert int(ssn[16]) % 2 == 0
        # Males have odd number at index 17
        ssn = self.fake.ssn(gender="M")
        assert int(ssn[16]) % 2 == 1

    def test_zh_CN_ssn_invalid_area_code_passed(self):
        ssn = self.fake.ssn(area_code=12)
        assert int(ssn[0:6]) > 0

        ssn = self.fake.ssn(area_code={})
        assert int(ssn[0:6]) > 0

        ssn = self.fake.ssn(area_code=[])
        assert int(ssn[0:6]) > 0

        ssn = self.fake.ssn(area_code=None)
        assert int(ssn[0:6]) > 0

        ssn = self.fake.ssn()
        assert int(ssn[0:6]) > 0

    def test_zh_CN_ssn_area_code_passed(self):
        #
        ssn = self.fake.ssn(area_code="654225")
        assert int(ssn[0:6]) == 654225

        ssn = self.fake.ssn(area_code="820000")
        assert int(ssn[0:6]) == 820000

        ssn = self.fake.ssn(area_code="830000")
        assert int(ssn[0:6]) == 830000


class TestRoRO(unittest.TestCase):
    """Tests SSN in the ro_RO locale"""

    def setUp(self):
        self.fake = Faker("ro_RO")
        Faker.seed(0)

    def test_ssn_checksum(self):
        assert ro_ssn_checksum("188050510739") == 0
        assert ro_ssn_checksum("181111512587") == 1
        assert ro_ssn_checksum("190123152499") == 2
        assert ro_ssn_checksum("601100452314") == 3
        assert ro_ssn_checksum("296072904713") == 4
        assert ro_ssn_checksum("601100452314") == 3
        assert ro_ssn_checksum("192080516368") == 6
        assert ro_ssn_checksum("602041144519") == 7
        assert ro_ssn_checksum("197061731387") == 8
        assert ro_ssn_checksum("294112120140") == 9

    def test_ssn(self):
        for _ in range(100):
            assert re.search(r"^\d{13}$", self.fake.ssn())

    def test_vat_checksum(self):
        assert ro_vat_checksum("1") == 9
        assert ro_vat_checksum("41") == 8
        assert ro_vat_checksum("181") == 2
        assert ro_vat_checksum("82421") == 5
        assert ro_vat_checksum("424694") == 7
        assert ro_vat_checksum("3918774") == 6
        assert ro_vat_checksum("99380784") == 1
        assert ro_vat_checksum("971775895") == 8

    def test_vat_id(self):
        for _ in range(100):
            vat = self.fake.vat_id().replace("RO", "")
            assert vat.isdigit()
            assert len(vat) >= 2 and len(vat) <= 10


class TestAzAz(unittest.TestCase):
    num_sample_runs = 10

    def setUp(self):
        self.fake = Faker("az_AZ")
        self.samples = [self.fake.ssn() for _ in range(self.num_sample_runs)]
        Faker.seed(0)

    def check_length(self):
        for sample in self.samples:
            assert len(sample) == 7


class TestLvLV(unittest.TestCase):
    num_sample_runs = 10

    def setUp(self):
        self.fake = Faker("lv_LV")
        Faker.seed(0)
        self.samples = [self.fake.ssn() for _ in range(self.num_sample_runs)]
        self.provider = lv_Provider

    def test_century_code(self):
        assert self.provider._get_century_code(1900) == 1
        assert self.provider._get_century_code(1999) == 1
        assert self.provider._get_century_code(2000) == 2
        assert self.provider._get_century_code(2999) == 2
        assert self.provider._get_century_code(1800) == 0
        assert self.provider._get_century_code(1899) == 0
        with pytest.raises(ValueError):
            self.provider._get_century_code(1799)
        with pytest.raises(ValueError):
            self.provider._get_century_code(3000)

    def test_ssn_sanity(self):
        for age in range(100):
            self.fake.ssn(min_age=age, max_age=age + 1)

    def check_length(self):
        for sample in self.samples:
            assert len(sample) == 12

    def test_vat_id(self):
        for _ in range(100):
            assert re.search(r"^LV\d{11}$", self.fake.vat_id())


class TestZhTW(unittest.TestCase):
    num_sample_runs = 10

    def setUp(self):
        self.fake = Faker("zh_TW")
        Faker.seed(0)
        self.samples = [self.fake.ssn() for _ in range(self.num_sample_runs)]

    def test_length(self):
        for sample in self.samples:
            assert len(sample) == 10

    def test_gender(self):
        """only '1' and '2' are allowed in the second char"""
        for sample in self.samples:
            assert sample[1] == "1" or sample[1] == "2"

    def test_checksum(self):
        for sample in self.samples:
            assert tw_checksum(sample) % 10 == 0


class TestUkUA(unittest.TestCase):
    def setUp(self):
        self.fake = Faker("uk_Ua")
        Faker.seed(0)
        self.provider = uk_Provider

    def test_ssn_len(self):
        assert len(self.fake.ssn()) == 10

    def test_start_ssn(self):
        assert self.fake.ssn("21-06-1994")[:5] == "34505"

    def test_ssn_gender(self):
        m = self.fake.ssn(gender="M")
        w = self.fake.ssn(gender="F")
        assert int(m[8]) % 2 != 0, "Must be odd for men"
        assert int(w[8]) % 2 == 0, "Must be even for women"

    def test_incorrect_birthday(self):
        with pytest.raises(ValueError):
            self.fake.ssn(birthday="1994-06-01")

    def test_incorrect_gender(self):
        with pytest.raises(ValueError):
            self.fake.ssn(gender="f")



from .. import Provider as DateTimeProvider


class Provider(DateTimeProvider):
    DAY_NAMES = {
        "0": "dimanche",
        "1": "lundi",
        "2": "mardi",
        "3": "mercredi",
        "4": "jeudi",
        "5": "vendredi",
        "6": "samedi",
    }
    MONTH_NAMES = {
        "01": "Janvier",
        "02": "Février",
        "03": "Mars",
        "04": "Avril",
        "05": "Mai",
        "06": "Juin",
        "07": "Juillet",
        "08": "Août",
        "09": "Septembre",
        "10": "Octobre",
        "11": "Novembre",
        "12": "Décembre",
    }

    def day_of_week(self) -> str:
        day = self.date("%w")
        return self.DAY_NAMES[day]

    def month_name(self) -> str:
        month = self.month()
        return self.MONTH_NAMES[month]



import random

from faker.utils.checksums import calculate_luhn

from .. import Provider as BaseProvider


def tin_checksum(tin: str) -> int:
    """
    Calculates the checksum (last) digit of Greek TINs given the rest
    :param tin: first 8 digits of a Greek TIN
    :return: calculated checksum digit
    """

    tin_list = [int(i) for i in list(tin)]
    return (
        (
            (tin_list[0] * 256)
            + (tin_list[1] * 128)
            + (tin_list[2] * 64)
            + (tin_list[3] * 32)
            + (tin_list[4] * 16)
            + (tin_list[5] * 8)
            + (tin_list[6] * 4)
            + (tin_list[7] * 2)
        )
        % 11
    ) % 10


class Provider(BaseProvider):
    """
    A Faker provider for Greek identification numbers
    """

    police_id_format = "??######"

    # TIN checksum algo sourced from here
    # http://epixeirisi.gr/%CE%9A%CE%A1%CE%99%CE%A3%CE%99%CE%9C%CE%91-%CE%98%CE%95%CE%9C%CE%91%CE%A4%CE%91-%CE%A6%CE%9F%CE%A1%CE%9F%CE%9B%CE%9F%CE%93%CE%99%CE%91%CE%A3-%CE%9A%CE%91%CE%99-%CE%9B%CE%9F%CE%93%CE%99%CE%A3%CE%A4%CE%99%CE%9A%CE%97%CE%A3/23791/%CE%91%CF%81%CE%B9%CE%B8%CE%BC%CF%8C%CF%82-%CE%A6%CE%BF%CF%81%CE%BF%CE%BB%CE%BF%CE%B3%CE%B9%CE%BA%CE%BF%CF%8D-%CE%9C%CE%B7%CF%84%CF%81%CF%8E%CE%BF%CF%85
    def vat_id(self, prefix: bool = True) -> str:
        """
        Generates random Greek VAT IDs (business TINs)
        :param prefix: boolean option to use EU format ("EL") prefix
        :return: a random Greek VAT ID
        """

        vat_id = "EL" if prefix else ""
        vat_id_starting_numbers = ("7", "8", "9", "0")
        vat_id = vat_id + random.choice(vat_id_starting_numbers) + self.numerify("#######")
        return vat_id + str(tin_checksum(vat_id[2:] if prefix else vat_id))

    def tin(self) -> str:
        """
        Generates random Greek personal TINs
        :return: a random Greek personal TIN
        """

        vat_id_starting_numbers = ("1", "2", "3", "4")
        vat_id = random.choice(vat_id_starting_numbers) + self.numerify("#######")
        return vat_id + str(tin_checksum(vat_id))

    # Uses Luhn checksum according to this
    # https://dotnetadventures.wordpress.com/2012/12/13/c-%CE%AD%CE%BB%CE%B5%CE%B3%CF%87%CE%BF%CF%82-%CE%BF%CF%81%CE%B8%CF%8C%CF%84%CE%B7%CF%84%CE%B1%CF%82-%CE%B1-%CE%BC-%CE%BA-%CE%B1-includes-python-version/
    def ssn(self) -> str:
        """
        Generates random Greek social security number (AMKA)
        :return: a random Greek social security number
        """

        ssn = self.generator.date(pattern="%d%m%y") + self.numerify("####")
        return ssn + str(calculate_luhn(ssn))

    # Valid format accd to ΥΑ 3021/19/53/2005 - FΕΚ 1440/Β'/18.10.2005
    # http://www.dsanet.gr/Epikairothta/Nomothesia/ya3021_19_05.htm
    def police_id(self) -> str:
        """
        Generates random Greek identity card (aka police-issued identification card) numbers
        :return: a random Greek identity card number
        """

        return self.bothify(
            self.police_id_format,
            letters="ΑΒΕΖΗΙΚΜΝΟΡΤΥΧ",
        )



import unittest

from faker import Faker


class TestGlobal(unittest.TestCase):
    """Test emoji provider methods"""

    def setUp(self):
        self.fake = Faker()  # No locale specified, gets global for this provider
        Faker.seed(0)

    def test_emoji(self):
        emoji = self.fake.emoji()
        assert isinstance(emoji, str)



from .. import Provider as PersonProvider


class Provider(PersonProvider):
    formats_male = (
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}",
        "{{first_name_male}} {{last_name}}-{{last_name}}",
        "{{prefix_male}} {{first_name_male}} {{last_name}}",
    )

    formats_female = (
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}",
        "{{first_name_female}} {{last_name}}-{{last_name}}",
        "{{prefix_female}} {{first_name_female}} {{last_name}}",
    )

    formats = formats_male + formats_female

    # source: https://www.data.gv.at/katalog/dataset/87fc82a0-0042-49c8-b6f9-2602cd3dc17a
    first_names_male = (
        "Aaron",
        "Adam",
        "Adrian",
        "Adriano",
        "Alan",
        "Aleksander",
        "Alex",
        "Alexandar",
        "Alexander",
        "Andreas",
        "Andrej",
        "Angelo",
        "Anton",
        "Antonio",
        "Antonius",
        "Arda",
        "Armin",
        "Aron",
        "Arthur",
        "Aurelio",
        "Axel",
        "Bastian",
        "Ben",
        "Benedict",
        "Benedikt",
        "Beni",
        "Benjamin",
        "Bernhard",
        "Boris",
        "Bruno",
        "Calvin",
        "Carl",
        "Carlo",
        "Chris",
        "Christian",
        "Christoph",
        "Christopher",
        "Clemens",
        "Constantin",
        "Cornelius",
        "Cristiano",
        "Damian",
        "Daniel",
        "Danilo",
        "Dario",
        "Darius",
        "Darko",
        "David",
        "Dennis",
        "Dominik",
        "Eduard",
        "Elias",
        "Elyas",
        "Emanuel",
        "Emil",
        "Emilian",
        "Emmanuel",
        "Eric",
        "Erik",
        "Erwin",
        "Fabian",
        "Fabio",
        "Felix",
        "Ferdinand",
        "Fernando",
        "Filip",
        "Finn",
        "Florentin",
        "Florian",
        "Florin",
        "Franz",
        "Frederik",
        "Fridolin",
        "Friedrich",
        "Gabriel",
        "Georg",
        "Gregor",
        "Gustav",
        "Heinrich",
        "Henri",
        "Henrik",
        "Henry",
        "Hubert",
        "Hugo",
        "Igor",
        "Ilias",
        "Isa",
        "Ismail",
        "Jacob",
        "Jakob",
        "James",
        "Jamie",
        "Jan",
        "Jannik",
        "Jannis",
        "Jasper",
        "Joel",
        "Johann",
        "Johannes",
        "John",
        "Jonas",
        "Jonathan",
        "Josef",
        "Joseph",
        "Joshua",
        "Julian",
        "Julius",
        "Justin",
        "Justus",
        "Kai",
        "Karim",
        "Karl",
        "Kevin",
        "Kilian",
        "Konrad",
        "Konstantin",
        "Kristian",
        "Lars",
        "Laurenz",
        "Laurin",
        "Lean",
        "Leander",
        "Lennard",
        "Lennart",
        "Leo",
        "Leon",
        "Leonard",
        "Leonardo",
        "Leonhard",
        "Leopold",
        "Levi",
        "Liam",
        "Lino",
        "Linus",
        "Lionel",
        "Lorenz",
        "Lorenzo",
        "Louis",
        "Luca",
        "Lucas",
        "Luis",
        "Luka",
        "Lukas",
        "Maksim",
        "Manuel",
        "Marc",
        "Marcel",
        "Marco",
        "Marcus",
        "Mario",
        "Marius",
        "Mark",
        "Marko",
        "Markus",
        "Martin",
        "Marvin",
        "Mateo",
        "Matheo",
        "Mathias",
        "Matteo",
        "Matthias",
        "Maurice",
        "Max",
        "Maximilian",
        "Merlin",
        "Mert",
        "Michael",
        "Mika",
        "Mike",
        "Milan",
        "Milo",
        "Moritz",
        "Natan",
        "Nathan",
        "Nicholas",
        "Nick",
        "Nico",
        "Nicolai",
        "Nicolas",
        "Niklas",
        "Niko",
        "Nikola",
        "Nikolai",
        "Nikolas",
        "Nikolaus",
        "Nils",
        "Nino",
        "Noah",
        "Noel",
        "Oliver",
        "Oscar",
        "Oskar",
        "Pascal",
        "Patrick",
        "Patrik",
        "Paul",
        "Peter",
        "Philip",
        "Philipp",
        "Phillip",
        "Raffael",
        "Ralph",
        "Raphael",
        "Rene",
        "Ricardo",
        "Richard",
        "Robert",
        "Robin",
        "Roman",
        "Ruben",
        "Sam",
        "Samuel",
        "Sandro",
        "Sascha",
        "Sebastian",
        "Severin",
        "Simon",
        "Stefan",
        "Stephan",
        "Steven",
        "Sven",
        "Teodor",
        "Theo",
        "Theodor",
        "Thomas",
        "Tim",
        "Timo",
        "Timon",
        "Tobias",
        "Tom",
        "Tristan",
        "Valentin",
        "Valentino",
        "Victor",
        "Viktor",
        "Vincent",
        "Vito",
        "William",
        "Xavier",
    )

    # source: https://www.data.gv.at/katalog/dataset/87fc82a0-0042-49c8-b6f9-2602cd3dc17a
    first_names_female = (
        "Adelina",
        "Adriana",
        "Ajna",
        "Alara",
        "Aleksandra",
        "Alena",
        "Alexa",
        "Alexandra",
        "Alexia",
        "Alice",
        "Alma",
        "Amanda",
        "Amelia",
        "Amelie",
        "Anabella",
        "Anastasia",
        "Andjela",
        "Andjelina",
        "Andrea",
        "Angela",
        "Angelika",
        "Angelina",
        "Anika",
        "Anita",
        "Anja",
        "Anna",
        "Anna-Lena",
        "Anna-Maria",
        "Annabell",
        "Annabella",
        "Annabelle",
        "Annalena",
        "Anne",
        "Annika",
        "Antonella",
        "Antonia",
        "Ariana",
        "Ariane",
        "Aurelia",
        "Aurora",
        "Ava",
        "Aylin",
        "Barbara",
        "Beatrice",
        "Bernadette",
        "Berra",
        "Bianca",
        "Carina",
        "Carla",
        "Carlotta",
        "Carolina",
        "Caroline",
        "Catharina",
        "Cecilia",
        "Charlotte",
        "Christina",
        "Christine",
        "Claire",
        "Clara",
        "Clarissa",
        "Claudia",
        "Constanze",
        "Cristina",
        "Dana",
        "Daniela",
        "Denise",
        "Diana",
        "Dilara",
        "Domenica",
        "Dora",
        "Eda",
        "Edda",
        "Ela",
        "Elena",
        "Eleonora",
        "Elina",
        "Elisa",
        "Elisabeth",
        "Ella",
        "Ellie",
        "Elma",
        "Elona",
        "Elsa",
        "Elvira",
        "Emanuela",
        "Emely",
        "Emilia",
        "Emilie",
        "Emilija",
        "Emma",
        "Erina",
        "Estelle",
        "Esther",
        "Eva",
        "Evelyn",
        "Felicitas",
        "Fiona",
        "Florentina",
        "Francesca",
        "Franziska",
        "Frida",
        "Gabriela",
        "Gloria",
        "Hanna",
        "Hannah",
        "Heidi",
        "Helena",
        "Helene",
        "Ina",
        "Ines",
        "Irina",
        "Iris",
        "Irma",
        "Isabel",
        "Isabell",
        "Isabella",
        "Isabelle",
        "Jana",
        "Janine",
        "Jasmina",
        "Jasmine",
        "Jennifer",
        "Jessica",
        "Johanna",
        "Josefine",
        "Jovana",
        "Julia",
        "Juliana",
        "Juliane",
        "Julijana",
        "Juna",
        "Kalina",
        "Karina",
        "Karla",
        "Karolina",
        "Karoline",
        "Katarina",
        "Katharina",
        "Katja",
        "Kerstin",
        "Klara",
        "Kristina",
        "Kyra",
        "Laetitia",
        "Laila",
        "Lana",
        "Lara",
        "Lara-Sophie",
        "Larissa",
        "Laura",
        "Laureen",
        "Lea",
        "Lea-Sophie",
        "Leah",
        "Leandra",
        "Lena",
        "Leni",
        "Leona",
        "Leoni",
        "Leonie",
        "Leonora",
        "Leontina",
        "Leticia",
        "Leyla",
        "Lia",
        "Lilia",
        "Lilian",
        "Liliana",
        "Liliane",
        "Lilli",
        "Lilly",
        "Lily",
        "Lina",
        "Linda",
        "Linnea",
        "Lisa",
        "Lisa-Marie",
        "Lola",
        "Lora",
        "Lorena",
        "Lotta",
        "Lotte",
        "Louisa",
        "Louise",
        "Luana",
        "Lucia",
        "Lucie",
        "Luisa",
        "Luise",
        "Luna",
        "Lydia",
        "Madeleine",
        "Magdalena",
        "Maida",
        "Maja",
        "Malena",
        "Manuela",
        "Mara",
        "Maria",
        "Mariam",
        "Mariana",
        "Marie",
        "Marie-Louise",
        "Marie-Sophie",
        "Mariella",
        "Marijana",
        "Marina",
        "Marissa",
        "Marlene",
        "Marta",
        "Martha",
        "Martina",
        "Maryam",
        "Mathilda",
        "Matilda",
        "Maya",
        "Melanie",
        "Melek",
        "Melina",
        "Melisa",
        "Melissa",
        "Mia",
        "Michaela",
        "Michelle",
        "Mila",
        "Milica",
        "Mina",
        "Mira",
        "Miriam",
        "Mona",
        "Nadia",
        "Nadin",
        "Nadine",
        "Nadja",
        "Naomi",
        "Natalia",
        "Natalie",
        "Natascha",
        "Nathalie",
        "Nela",
        "Nele",
        "Nelly",
        "Nicola",
        "Nicole",
        "Nika",
        "Nikita",
        "Nikola",
        "Nikolina",
        "Nina",
        "Nisa",
        "Nora",
        "Norah",
        "Olivia",
        "Patricia",
        "Paula",
        "Paulina",
        "Pauline",
        "Petra",
        "Philippa",
        "Pia",
        "Rachel",
        "Raffaela",
        "Rana",
        "Rayana",
        "Rebecca",
        "Rita",
        "Romy",
        "Ronja",
        "Ronya",
        "Rosa",
        "Rosalie",
        "Ruth",
        "Sabine",
        "Sabrina",
        "Sahra",
        "Salma",
        "Sandra",
        "Sara",
        "Sarah",
        "Selena",
        "Selin",
        "Selina",
        "Selma",
        "Sena",
        "Siena",
        "Sigrid",
        "Sofia",
        "Sofie",
        "Sofija",
        "Sonja",
        "Sophia",
        "Sophie",
        "Sophie-Marie",
        "Soraya",
        "Stefanie",
        "Stella",
        "Stephanie",
        "Tamara",
        "Tanja",
        "Tea",
        "Theodora",
        "Theresa",
        "Therese",
        "Tiffany",
        "Tina",
        "Valentina",
        "Vanessa",
        "Vera",
        "Verena",
        "Veronika",
        "Victoria",
        "Viktoria",
        "Viola",
        "Violetta",
        "Vivian",
        "Yasmina",
        "Ylvie",
        "Yvonne",
        "Zara",
        "Zoe",
        "Zoey",
    )

    first_names = first_names_male + first_names_female

    # about 1000 of the most popular Austrian surnames
    # https://de.wiktionary.org/wiki/Verzeichnis:Deutsch/Namen/die_h%C3%A4ufigsten_Nachnamen_%C3%96sterreichs
    last_names = (
        "Achleitner",
        "Ackerl",
        "Adam",
        "Adler",
        "Aichholzer",
        "Aichinger",
        "Aigner",
        "Albrecht",
        "Altmann",
        "Amann",
        "Amon",
        "Anderl",
        "Angerer",
        "Arnold",
        "Artner",
        "Aschauer",
        "Auer",
        "Augustin",
        "Auinger",
        "Bacher",
        "Bachinger",
        "Bachler",
        "Bachmann",
        "Bader",
        "Baier",
        "Baldauf",
        "Barth",
        "Bartl",
        "Bauer",
        "Baumann",
        "Baumgartner",
        "Bayer",
        "Beck",
        "Becker",
        "Beer",
        "Berchtold",
        "Berger",
        "Bergmann",
        "Berner",
        "Bernhard",
        "Berthold",
        "Bichler",
        "Biedermann",
        "Binder",
        "Bischof",
        "Bitschnau",
        "Bittner",
        "Blauensteiner",
        "Blum",
        "Blümel",
        "Bock",
        "Bodner",
        "Bogner",
        "Brandl",
        "Brandner",
        "Brandstetter",
        "Brandstätter",
        "Brandtner",
        "Braun",
        "Brenner",
        "Breuer",
        "Bruckner",
        "Brugger",
        "Brunner",
        "Bräuer",
        "Buchberger",
        "Buchegger",
        "Bucher",
        "Buchinger",
        "Buchner",
        "Burger",
        "Burgstaller",
        "Burtscher",
        "Bäck",
        "Böck",
        "Böhler",
        "Böhm",
        "Bösch",
        "Bürger",
        "Dallinger",
        "Dangl",
        "Danner",
        "Danninger",
        "Decker",
        "Dengg",
        "Denk",
        "Deutschmann",
        "Dietl",
        "Dietrich",
        "Dirnberger",
        "Dittrich",
        "Dobler",
        "Doppler",
        "Dorfer",
        "Dorn",
        "Dorner",
        "Draxler",
        "Dünser",
        "Eberhard",
        "Eberharter",
        "Eberl",
        "Ebner",
        "Ecker",
        "Eder",
        "Edlinger",
        "Egger",
        "Eibl",
        "Eichberger",
        "Eichhorn",
        "Eichinger",
        "Eisl",
        "Eisner",
        "Eller",
        "Ender",
        "Engel",
        "Engl",
        "Enzinger",
        "Erber",
        "Erhart",
        "Erlacher",
        "Erler",
        "Ernst",
        "Ertl",
        "Fabian",
        "Falkner",
        "Fankhauser",
        "Farkas",
        "Fasching",
        "Fehringer",
        "Feichtenschlager",
        "Feichter",
        "Feichtinger",
        "Feichtner",
        "Feigl",
        "Felber",
        "Felder",
        "Fellinger",
        "Fellner",
        "Fercher",
        "Ferstl",
        "Fichtinger",
        "Fiedler",
        "Fink",
        "Fischer",
        "Fitz",
        "Fleck",
        "Fleischhacker",
        "Fleischmann",
        "Foidl",
        "Forster",
        "Forstner",
        "Frank",
        "Franz",
        "Freitag",
        "Freudenthaler",
        "Frey",
        "Frick",
        "Friedl",
        "Friedrich",
        "Frisch",
        "Fritsch",
        "Fritz",
        "Froschauer",
        "Fröhlich",
        "Fröschl",
        "Frühwirth",
        "Fuchs",
        "Fuhrmann",
        "Füreder",
        "Fürst",
        "Gabriel",
        "Gahleitner",
        "Galler",
        "Gamsjäger",
        "Gangl",
        "Gartner",
        "Gasser",
        "Gassner",
        "Gattringer",
        "Geier",
        "Geiger",
        "Geisler",
        "Geyer",
        "Gindl",
        "Glaser",
        "Glatz",
        "Glück",
        "Gmeiner",
        "Gollner",
        "Gosch",
        "Grabher",
        "Grabner",
        "Graf",
        "Grasser",
        "Grassl",
        "Gratz",
        "Gratzer",
        "Gratzl",
        "Greiner",
        "Griesser",
        "Grill",
        "Gritsch",
        "Gross",
        "Groß",
        "Gruber",
        "Grundner",
        "Grünberger",
        "Grüner",
        "Grünwald",
        "Gschaider",
        "Gschwandtner",
        "Gstrein",
        "Guggenberger",
        "Gutmann",
        "Gärtner",
        "Göschl",
        "Götz",
        "Günther",
        "Haas",
        "Haberl",
        "Hacker",
        "Hackl",
        "Haderer",
        "Hafner",
        "Hagen",
        "Hager",
        "Hahn",
        "Haid",
        "Haiden",
        "Haider",
        "Haidinger",
        "Haindl",
        "Hainzl",
        "Haller",
        "Hammer",
        "Hammerer",
        "Hammerl",
        "Handl",
        "Handler",
        "Haring",
        "Harrer",
        "Hartl",
        "Hartmann",
        "Haslauer",
        "Haslinger",
        "Hattinger",
        "Hauer",
        "Haumer",
        "Hausberger",
        "Hauser",
        "Hebenstreit",
        "Hechenberger",
        "Heger",
        "Heigl",
        "Heim",
        "Heindl",
        "Heinrich",
        "Heinz",
        "Heinzl",
        "Heiss",
        "Heissenberger",
        "Held",
        "Hell",
        "Heller",
        "Helm",
        "Hemetsberger",
        "Herbst",
        "Hermann",
        "Herrmann",
        "Herzog",
        "Himmelbauer",
        "Hinterberger",
        "Hinteregger",
        "Hinterleitner",
        "Hirsch",
        "Hirschmann",
        "Hochleitner",
        "Hochreiter",
        "Hofbauer",
        "Hofer",
        "Hoffmann",
        "Hofinger",
        "Hofmann",
        "Hofmeister",
        "Hofstetter",
        "Hofstätter",
        "Holl",
        "Hollaus",
        "Holler",
        "Holzer",
        "Holzinger",
        "Holzknecht",
        "Holzmann",
        "Horak",
        "Horn",
        "Hosp",
        "Huber",
        "Hubmann",
        "Huemer",
        "Hufnagl",
        "Humer",
        "Hummel",
        "Hummer",
        "Huter",
        "Hutter",
        "Hutterer",
        "Hämmerle",
        "Häusler",
        "Hödl",
        "Höfer",
        "Höfler",
        "Höglinger",
        "Höller",
        "Hölzl",
        "Hörl",
        "Hörmann",
        "Hübner",
        "Hütter",
        "Jahn",
        "Jandl",
        "Janisch",
        "Jank",
        "Jauk",
        "Jenewein",
        "Jost",
        "Jovanovic",
        "Juen",
        "Jung",
        "Jungwirth",
        "Jäger",
        "Jöbstl",
        "Kager",
        "Kahr",
        "Kain",
        "Kaindl",
        "Kainz",
        "Kaiser",
        "Kalcher",
        "Kaltenbrunner",
        "Kaltenböck",
        "Kaltenegger",
        "Kammerer",
        "Kapeller",
        "Kappel",
        "Kargl",
        "Karl",
        "Karner",
        "Karrer",
        "Kaspar",
        "Kasper",
        "Kastner",
        "Kaufmann",
        "Keller",
        "Kellner",
        "Keplinger",
        "Kern",
        "Kerschbaum",
        "Kerschbaumer",
        "Kessler",
        "Kirchmair",
        "Kirchner",
        "Kirschner",
        "Kiss",
        "Kitzler",
        "Klammer",
        "Klaus",
        "Klausner",
        "Klein",
        "Klement",
        "Klinger",
        "Klingler",
        "Klocker",
        "Kloiber",
        "Klotz",
        "Klug",
        "Knapp",
        "Knaus",
        "Knoll",
        "Kober",
        "Koch",
        "Kocher",
        "Kofler",
        "Kogler",
        "Kohl",
        "Kohler",
        "Kolar",
        "Kolb",
        "Koller",
        "Kollmann",
        "Kolm",
        "Konrad",
        "Kopf",
        "Kopp",
        "Koppensteiner",
        "Kraft",
        "Krainer",
        "Krainz",
        "Kral",
        "Krall",
        "Kramer",
        "Krammer",
        "Kratzer",
        "Kraus",
        "Kraxner",
        "Kreidl",
        "Kreiner",
        "Kremser",
        "Krenn",
        "Kreuzer",
        "Kriegl",
        "Kronberger",
        "Kronsteiner",
        "Krug",
        "Kröll",
        "Kucera",
        "Kugler",
        "Kuhn",
        "Kummer",
        "Kunz",
        "Kurz",
        "Kurzmann",
        "Käfer",
        "Köberl",
        "Köck",
        "Köhler",
        "Kölbl",
        "Köll",
        "König",
        "Kössler",
        "Lackner",
        "Ladner",
        "Lagler",
        "Laimer",
        "Lammer",
        "Lampert",
        "Lampl",
        "Lamprecht",
        "Landl",
        "Lang",
        "Langer",
        "Larcher",
        "Lassnig",
        "Leber",
        "Lechner",
        "Lederer",
        "Leeb",
        "Lehner",
        "Leibetseder",
        "Leitgeb",
        "Leithner",
        "Leitner",
        "Lengauer",
        "Lenz",
        "Leonhartsberger",
        "Leopold",
        "Lerch",
        "Lercher",
        "Lettner",
        "Leutgeb",
        "Lichtenegger",
        "Linder",
        "Lindinger",
        "Lindner",
        "Lindorfer",
        "Lintner",
        "Lipp",
        "List",
        "Loibl",
        "Loidl",
        "Lorenz",
        "Ludwig",
        "Luger",
        "Luttenberger",
        "Lutz",
        "Löffler",
        "Macher",
        "Mader",
        "Maier",
        "Maierhofer",
        "Mair",
        "Mairhofer",
        "Mandl",
        "Mann",
        "Margreiter",
        "Maringer",
        "Mark",
        "Markl",
        "Marte",
        "Martin",
        "Marx",
        "Mathis",
        "Maurer",
        "Mayer",
        "Mayerhofer",
        "Mayr",
        "Mayrhofer",
        "Meier",
        "Meindl",
        "Meister",
        "Meixner",
        "Messner",
        "Metzler",
        "Meusburger",
        "Meyer",
        "Mitter",
        "Mitteregger",
        "Mitterer",
        "Mitterlehner",
        "Mittermayr",
        "Mohr",
        "Moosbrugger",
        "Moritz",
        "Moser",
        "Muhr",
        "Mörth",
        "Mühlbacher",
        "Mühlberger",
        "Mühlböck",
        "Müller",
        "Müllner",
        "Nagel",
        "Nagele",
        "Nagl",
        "Nemeth",
        "Neubacher",
        "Neubauer",
        "Neugebauer",
        "Neuhauser",
        "Neuhold",
        "Neulinger",
        "Neumann",
        "Neumayer",
        "Neumayr",
        "Neumeister",
        "Neumüller",
        "Neuner",
        "Neureiter",
        "Neuwirth",
        "Niederl",
        "Nowak",
        "Nussbaumer",
        "Nußbaumer",
        "Nöbauer",
        "Oberhauser",
        "Oberhofer",
        "Oberleitner",
        "Obermayr",
        "Obermüller",
        "Oberndorfer",
        "Ofner",
        "Ortner",
        "Ostermann",
        "Oswald",
        "Ott",
        "Pacher",
        "Pachler",
        "Paier",
        "Pammer",
        "Parzer",
        "Pauer",
        "Paul",
        "Paulitsch",
        "Payer",
        "Peer",
        "Peham",
        "Pendl",
        "Penz",
        "Perner",
        "Pertl",
        "Pesendorfer",
        "Peter",
        "Petz",
        "Pfeffer",
        "Pfeifer",
        "Pfeiffer",
        "Pfister",
        "Pfleger",
        "Philipp",
        "Pichler",
        "Pieber",
        "Pilz",
        "Pinter",
        "Pircher",
        "Pirker",
        "Plank",
        "Plattner",
        "Platzer",
        "Pock",
        "Pohl",
        "Pointner",
        "Pokorny",
        "Pollak",
        "Polzer",
        "Posch",
        "Postl",
        "Prager",
        "Prantl",
        "Praxmarer",
        "Prem",
        "Prenner",
        "Prinz",
        "Probst",
        "Prohaska",
        "Pröll",
        "Pucher",
        "Puchner",
        "Puntigam",
        "Punz",
        "Putz",
        "Pöll",
        "Pölzl",
        "Pöschl",
        "Pühringer",
        "Raab",
        "Rabitsch",
        "Rabl",
        "Radl",
        "Rainer",
        "Ramsauer",
        "Rath",
        "Rauch",
        "Rausch",
        "Rauscher",
        "Rauter",
        "Rechberger",
        "Redl",
        "Reich",
        "Reichel",
        "Reicher",
        "Reichl",
        "Reichmann",
        "Reif",
        "Reinbacher",
        "Reindl",
        "Reiner",
        "Reinisch",
        "Reinprecht",
        "Reinthaler",
        "Reischl",
        "Reisinger",
        "Reisner",
        "Reitbauer",
        "Reiter",
        "Reiterer",
        "Reithofer",
        "Reitinger",
        "Renner",
        "Resch",
        "Rettenbacher",
        "Richter",
        "Rieder",
        "Riedl",
        "Riedler",
        "Riedmann",
        "Rieger",
        "Riegler",
        "Riener",
        "Riepl",
        "Rieser",
        "Ringhofer",
        "Rinner",
        "Ritter",
        "Rohrer",
        "Rohrmoser",
        "Rosenberger",
        "Rosner",
        "Rossmann",
        "Roth",
        "Rottensteiner",
        "Rotter",
        "Rudolf",
        "Rupp",
        "Röck",
        "Rössler",
        "Sagmeister",
        "Sailer",
        "Salcher",
        "Salzer",
        "Salzmann",
        "Sammer",
        "Santner",
        "Sattler",
        "Sauer",
        "Schachinger",
        "Schachner",
        "Schaffer",
        "Schalk",
        "Schaller",
        "Schandl",
        "Schantl",
        "Scharf",
        "Scharinger",
        "Schartner",
        "Schatz",
        "Schatzl",
        "Schauer",
        "Scheer",
        "Scheiber",
        "Scheidl",
        "Schenk",
        "Scherer",
        "Scherr",
        "Scherz",
        "Scherzer",
        "Scheucher",
        "Schiefer",
        "Schiestl",
        "Schilcher",
        "Schiller",
        "Schimpl",
        "Schinagl",
        "Schindler",
        "Schinnerl",
        "Schlager",
        "Schlosser",
        "Schlögl",
        "Schmid",
        "Schmidinger",
        "Schmidl",
        "Schmidt",
        "Schmied",
        "Schmuck",
        "Schmölzer",
        "Schnabl",
        "Schneeberger",
        "Schneider",
        "Schober",
        "Scholz",
        "Schramm",
        "Schrammel",
        "Schranz",
        "Schreiber",
        "Schreiner",
        "Schrempf",
        "Schrenk",
        "Schrittwieser",
        "Schröder",
        "Schubert",
        "Schuh",
        "Schuler",
        "Schuller",
        "Schulz",
        "Schuster",
        "Schwab",
        "Schwaiger",
        "Schwaighofer",
        "Schwarz",
        "Schwarzinger",
        "Schwarzl",
        "Schweiger",
        "Schweighofer",
        "Schweitzer",
        "Schwendinger",
        "Schäfer",
        "Schöberl",
        "Schöffmann",
        "Schöller",
        "Schön",
        "Schönauer",
        "Schönberger",
        "Schöpf",
        "Schüller",
        "Schütz",
        "Seebacher",
        "Seidl",
        "Seifert",
        "Seiler",
        "Seiser",
        "Seitz",
        "Seiwald",
        "Sieber",
        "Sieberer",
        "Siegl",
        "Sigl",
        "Siller",
        "Simic",
        "Simon",
        "Singer",
        "Sommer",
        "Sonnleitner",
        "Sorger",
        "Sperl",
        "Spiegl",
        "Spindler",
        "Spitzer",
        "Spreitzer",
        "Springer",
        "Stadlbauer",
        "Stadler",
        "Stangl",
        "Stark",
        "Staudacher",
        "Staudinger",
        "Stecher",
        "Stefan",
        "Steger",
        "Steidl",
        "Steiger",
        "Steinacher",
        "Steinbacher",
        "Steinbauer",
        "Steinberger",
        "Steinböck",
        "Steindl",
        "Steiner",
        "Steininger",
        "Steinkellner",
        "Steinlechner",
        "Steinwender",
        "Stelzer",
        "Stelzl",
        "Stern",
        "Steurer",
        "Stiegler",
        "Stifter",
        "Stock",
        "Stocker",
        "Stockhammer",
        "Stockinger",
        "Stoiber",
        "Stolz",
        "Strasser",
        "Strauss",
        "Strauß",
        "Streicher",
        "Strobl",
        "Strohmaier",
        "Strohmayer",
        "Strohmeier",
        "Stummer",
        "Sturm",
        "Stöckl",
        "Stöger",
        "Stückler",
        "Stütz",
        "Sulzer",
        "Suppan",
        "Taferner",
        "Tanzer",
        "Tauber",
        "Taucher",
        "Teufl",
        "Thaler",
        "Thalhammer",
        "Thaller",
        "Thurner",
        "Tiefenbacher",
        "Tischler",
        "Toth",
        "Trattner",
        "Trauner",
        "Traxler",
        "Trimmel",
        "Trinkl",
        "Trummer",
        "Uhl",
        "Ullmann",
        "Ulrich",
        "Unger",
        "Unterberger",
        "Unterweger",
        "Urban",
        "Varga",
        "Veit",
        "Vogel",
        "Vogl",
        "Vogler",
        "Vogt",
        "Wachter",
        "Wagner",
        "Walch",
        "Walcher",
        "Walder",
        "Waldner",
        "Wallner",
        "Walser",
        "Walter",
        "Waltl",
        "Wandl",
        "Weber",
        "Wechselberger",
        "Wegscheider",
        "Weidinger",
        "Weigl",
        "Weinberger",
        "Weiser",
        "Weiss",
        "Weissenböck",
        "Weiß",
        "Wenger",
        "Weninger",
        "Wenzl",
        "Werner",
        "Widhalm",
        "Widmann",
        "Wiedner",
        "Wieland",
        "Wiener",
        "Wiesbauer",
        "Wieser",
        "Wiesinger",
        "Wiesner",
        "Wild",
        "Wilfinger",
        "Wilhelm",
        "Wimmer",
        "Windhager",
        "Windisch",
        "Winkler",
        "Winter",
        "Wirth",
        "Wittmann",
        "Wohlmuth",
        "Wolf",
        "Wurm",
        "Wurzer",
        "Wurzinger",
        "Wögerbauer",
        "Wöhrer",
        "Yilmaz",
        "Zach",
        "Zangerl",
        "Zauner",
        "Zechmeister",
        "Zechner",
        "Zehetner",
        "Zeiler",
        "Zeilinger",
        "Zeiner",
        "Zeller",
        "Zenz",
        "Zettl",
        "Ziegler",
        "Zimmermann",
        "Zotter",
        "Zöchling",
        "Zöhrer",
    )

    prefixes_male = (
        "Herr",
        "Dr.",
        "Ing.",
        "Dipl.-Ing.",
        "Prof.",
        "Univ.Prof.",
    )
    prefixes_female = (
        "Frau",
        "Dr.",
        "Ing.",
        "Dipl.-Ing.",
        "Prof.",
        "Univ.Prof.",
    )

    prefixes_male = (
        "Herr",
        "Dr.",
        "Ing.",
        "Dipl.-Ing.",
        "Prof.",
        "Univ.Prof.",
    )
    prefixes_female = (
        "Frau",
        "Dr.",
        "Ing.",
        "Dipl.-Ing.",
        "Prof.",
        "Univ.Prof.",
    )

    prefixes = ("Dr.", "Mag.", "Ing.", "Dipl.-Ing.", "Prof.", "Univ.Prof.")



import re
import string

from .. import Provider as AutomotiveProvider


class Provider(AutomotiveProvider):
    """Implement automotive provider for `nl_NL` locale.

    Sources:
    - https://en.wikipedia.org/wiki/Vehicle_registration_plates_of_the_Netherlands
    - https://www.cbs.nl/en-gb/figures/detail/82044eng

    .. |license_plate_car| replace::
       :meth:`license_plate_car() <faker.providers.automotive.nl_NL.Provider.license_plate_car>`

    .. |license_plate_motorbike| replace::
       :meth:`license_plate_motorbike() <faker.providers.automotive.nl_NL.Provider.license_plate_motorbike>`
    """

    # License formats for cars / other vehicles than motorbikes
    license_formats = (
        # Format 6
        "##-%?-??",
        # Format 7
        "##-%??-#",
        # Format 8
        "#-@??-##",
        # Format 9
        "%?-###-?",
        # Format 10
        "%-###-??",
    )

    # License formats for motorbikes.
    # According to CBS, approximately 10% of road vehicles in the Netherlands are motorbikes
    license_formats_motorbike = (
        "M?-??-##",
        "##-M?-??",
    )

    # Base first letters of format
    license_plate_prefix_letters = "BDFGHJKLNPRSTVXZ"

    # For Format 8 (9-XXX-99) "BDFGHJLNPR" are not used,
    # as to not clash with former export license plates
    license_plate_prefix_letters_format_8 = "KSTVXZ"

    def license_plate_motorbike(self) -> str:
        """Generate a license plate for motorbikes."""
        return self.bothify(
            self.random_element(self.license_formats_motorbike),
            letters=string.ascii_uppercase,
        )

    def license_plate_car(self) -> str:
        """Generate a license plate for cars."""
        # Replace % with license_plate_prefix_letters
        temp = re.sub(
            r"\%",
            self.random_element(self.license_plate_prefix_letters),
            self.random_element(self.license_formats),
        )

        # Replace @ with license_plate_prefix_letters_format_8
        temp = re.sub(r"\@", self.random_element(self.license_plate_prefix_letters_format_8), temp)

        return self.bothify(temp, letters=string.ascii_uppercase)

    def license_plate(self) -> str:
        """Generate a license plate.
        This method randomly chooses 10% between |license_plate_motorbike|
        or 90% |license_plate_car| to generate the result.
        """
        if self.generator.random.random() < 0.1:
            return self.license_plate_motorbike()
        return self.license_plate_car()



from .. import Provider as SsnProvider


class Provider(SsnProvider):
    def ssn(self) -> str:
        """
        Returns a 9 digits Dutch SSN called "burgerservicenummer (BSN)".

        the Dutch "burgerservicenummer (BSN)" needs to pass the "11-proef",
        which is a check digit approach; this function essentially reverses
        the checksum steps to create a random valid BSN (which is 9 digits).
        """

        # see http://nl.wikipedia.org/wiki/Burgerservicenummer (in Dutch)
        def _checksum(digits):
            factors = (9, 8, 7, 6, 5, 4, 3, 2, -1)
            s = 0
            for i in range(len(digits)):
                s += digits[i] * factors[i]
            return s

        while True:
            # create an array of first 8 elements initialized randomly
            digits = self.generator.random.sample(range(10), 8)
            # sum those 8 digits according to (part of) the "11-proef"
            s = _checksum(digits)
            # determine the last digit to make it qualify the test
            digits.append((s % 11) % 10)
            # repeat steps until it does qualify the test
            if 0 == (_checksum(digits) % 11):
                break

        # build the resulting BSN
        bsn = "".join([str(e) for e in digits])
        # finally return our random but valid BSN
        return bsn

    vat_id_formats = ("NL#########B##",)

    def vat_id(self) -> str:
        """
        http://ec.europa.eu/taxation_customs/vies/faq.html#item_11
        :return: A random Dutch VAT ID
        """
        return self.bothify(self.random_element(self.vat_id_formats))



from typing import OrderedDict

import pytest

from faker import Faker
from faker.providers import DynamicProvider


class TestDynamicProvider:
    def test_without_dynamic(self):
        faker = Faker()
        with pytest.raises(
            AttributeError,
            match="'Generator' object has no attribute 'medical_profession'",
        ):
            faker.medical_profession()

    def test_with_dynamic(self):
        faker = Faker()
        elements = ["dr.", "doctor", "nurse", "surgeon", "clerk"]
        provider_name = "medical_profession"

        medical_professions_provider = DynamicProvider(
            provider_name=provider_name,
            elements=elements,
        )

        faker.add_provider(medical_professions_provider)

        assert faker.medical_profession() in elements

    def test_dynamic_with_special_provider_name(self):
        elements = ["dr.", "doctor", "nurse", "surgeon", "clerk"]
        provider_name = "__special__"  # The provider name cannot start with __

        with pytest.raises(
            ValueError,
            match="Provider name cannot start with __ as it would be ignored by Faker",
        ):
            DynamicProvider(
                provider_name=provider_name,
                elements=elements,
            )

    def test_dynamic_with_empty_elements(self):
        elements = []
        provider_name = "my_provider"
        provider = DynamicProvider(
            provider_name=provider_name,
            elements=elements,
        )
        faker = Faker()
        faker.add_provider(provider)

        with pytest.raises(
            ValueError,
            match="Elements should be a list of values the provider samples from",
        ):
            faker.my_provider()

    def test_dynamic_add_element(self):
        elements = []
        provider_name = "my_provider"
        provider = DynamicProvider(
            provider_name=provider_name,
            elements=elements,
        )
        faker = Faker()
        faker.add_provider(provider)

        provider.add_element("one")
        provider.add_element("two")

        assert faker.my_provider() in ("one", "two")

    def test_weighted_dynamic_with_use_weighting(self):
        elements = OrderedDict([("A", 0.75), ("B", 0.25), ("C", 0.0)])
        provider_name = "my_provider"
        provider = DynamicProvider(
            provider_name=provider_name,
            elements=elements,
        )
        faker = Faker()
        faker.add_provider(provider)

        fake_data = [faker.my_provider(use_weighting=True) for _ in range(10_000)]

        for i in fake_data:
            assert i in {"A", "B"}

    def test_weighted_dynamic_without_use_weighting(self):
        elements = OrderedDict([("A", 0.75), ("B", 0.25), ("C", 0.0)])
        provider_name = "my_provider"
        provider = DynamicProvider(
            provider_name=provider_name,
            elements=elements,
        )
        faker = Faker()
        faker.add_provider(provider)

        fake_data = [faker.my_provider(use_weighting=False) for _ in range(10_000)]

        for i in fake_data:
            assert i in {"A", "B", "C"}



from .. import Provider as ColorProvider


class Provider(ColorProvider):
    """Implement color provider for ``sk_SK`` locale."""

    safe_colors = (
        "čierna",
        "gaštanová",
        "zelená",
        "námornícka",
        "olivová",
        "fialová",
        "zelenomodrá",
        "limetková",
        "modrá",
        "strieborná",
        "sivá",
        "žltá",
        "fuchsiová",
        "aquamarinová",
        "biela",
    )



from .. import Provider as AutomotiveProvider


class Provider(AutomotiveProvider):
    """Implement automotive provider for ``ar_BH`` locale.

    Source:

    - https://en.wikipedia.org/wiki/Vehicle_registration_plates_of_Bahrain
    """

    license_formats = ("######",)



from .. import Provider as CompanyProvider


class Provider(CompanyProvider):
    formats = (
        "{{last_name}} {{company_suffix}}",
        "{{last_name}} & {{last_name}} {{company_suffix}}",
        "{{last_name}} & Søn {{company_suffix}}",
    )

    company_suffixes = (
        "A/S",
        "ApS",
    )



from .. import Provider as BankProvider


class Provider(BankProvider):
    """Implement bank provider for ``sk_SK`` locale.

    https://www.mbank.cz/informace-k-produktum/info/ucty/cislo-uctu-iban.html
    """

    bban_format = "####################"
    country_code = "SK"



import pytest

from faker import Faker


class TestOptionalClass:
    def test_optional(self) -> None:
        fake = Faker()

        assert {fake.optional.boolean() for _ in range(10)} == {True, False, None}

    def test_optional_probability(self) -> None:
        """The probability is configurable."""
        fake = Faker()

        fake.optional.name(prob=0.1)

    def test_optional_arguments(self) -> None:
        """Other arguments are passed through to the function."""
        fake = Faker()

        fake.optional.pyint(1, 2, prob=0.4)

    def test_optional_valid_range(self) -> None:
        """Only probabilities in the range (0, 1]."""
        fake = Faker()

        with pytest.raises(ValueError, match=""):
            fake.optional.name(prob=0)

        with pytest.raises(ValueError, match=""):
            fake.optional.name(prob=1.1)

        with pytest.raises(ValueError, match=""):
            fake.optional.name(prob=-3)

    def test_functions_only(self):
        """Accessing non-functions through the `.optional` attribute
        will throw a TypeError."""

        fake = Faker()

        with pytest.raises(TypeError, match="Accessing non-functions through .optional is not supported."):
            fake.optional.locales



import re

from typing import Pattern

import pytest


class TestBarcodeProvider:
    """Test barcode provider methods"""

    num_samples = 1000
    ean8_pattern: Pattern = re.compile(r"\d{8}")
    ean13_pattern: Pattern = re.compile(r"\d{13}")

    def test_ean(self, faker, num_samples):
        for _ in range(num_samples):
            ean8 = faker.ean(8)
            ean13 = faker.ean(13)
            assert self.ean8_pattern.fullmatch(ean8)
            assert self.ean13_pattern.fullmatch(ean13)

            ean8_digits = [int(digit) for digit in ean8]
            ean13_digits = [int(digit) for digit in ean13]
            assert (sum(ean8_digits) + 2 * sum(ean8_digits[::2])) % 10 == 0
            assert (sum(ean13_digits) + 2 * sum(ean13_digits[1::2])) % 10 == 0

    def test_ean_bad_length(self, faker):
        bad_lengths = [size for size in range(1, 15) if size not in (8, 13)]
        for length in bad_lengths:
            with pytest.raises(AssertionError):
                faker.ean(length)

    def test_ean8(self, faker, num_samples):
        for _ in range(num_samples):
            ean8 = faker.ean8()
            assert self.ean8_pattern.fullmatch(ean8)

            # Included check digit must be correct
            ean8_digits = [int(digit) for digit in ean8]
            assert (sum(ean8_digits) + 2 * sum(ean8_digits[::2])) % 10 == 0

    def test_ean13(self, faker, num_samples):
        for _ in range(num_samples):
            ean13 = faker.ean13()
            assert self.ean13_pattern.fullmatch(ean13)

            # Included check digit must be correct
            ean13_digits = [int(digit) for digit in ean13]
            assert (sum(ean13_digits) + 2 * sum(ean13_digits[1::2])) % 10 == 0

    def test_ean13_no_leading_zero(self, faker, num_samples):
        for _ in range(num_samples):
            ean13 = faker.ean13(leading_zero=False)
            assert self.ean13_pattern.fullmatch(ean13)
            assert ean13[0] != "0"

            # Included check digit must be correct
            ean13_digits = [int(digit) for digit in ean13]
            assert (sum(ean13_digits) + 2 * sum(ean13_digits[1::2])) % 10 == 0

    def test_ean13_leading_zero(self, faker, num_samples):
        for _ in range(num_samples):
            ean13 = faker.ean13(leading_zero=True)
            assert self.ean13_pattern.fullmatch(ean13)
            assert ean13[0] == "0"

            # Included check digit must be correct
            ean13_digits = [int(digit) for digit in ean13]
            assert (sum(ean13_digits) + 2 * sum(ean13_digits[1::2])) % 10 == 0


@pytest.fixture(scope="class")
def provider_class(request):
    if hasattr(request.cls, "get_provider_class") and callable(request.cls.get_provider_class):
        _provider_class = request.cls.get_provider_class()
        if isinstance(_provider_class, type):
            return _provider_class
    raise NotImplementedError(
        f"Using the provider_class requires {request.cls.__name__}.get_provider_class() "
        "to be present, which has to return the Provider class it uses."
    )


@pytest.fixture()
def provider(faker, provider_class):
    return provider_class(faker)


class _LocaleCommonMixin:
    ean8_pattern: Pattern = re.compile(r"\d{8}")
    ean13_pattern: Pattern = re.compile(r"\d{13}")

    @staticmethod
    def assert_prefix(barcode_digits, prefixes):
        for prefix in prefixes:
            if all(a == b for a, b in zip(barcode_digits, map(int, prefix))):
                return
        str_barc = "".join(str(x) for x in barcode_digits)
        str_pref = ", ".join(map(lambda _prefix: "".join(str(x) for x in _prefix)), prefixes)
        raise AssertionError(f"{str_barc} doesn't match any of the prefixes: {str_pref}")

    def test_localized_ean(self, faker, num_samples, provider):
        for _ in range(num_samples):
            ean8 = faker.localized_ean(8)
            ean13 = faker.localized_ean(13)
            assert self.ean8_pattern.match(ean8)
            assert self.ean13_pattern.match(ean13)

            ean8_digits = [int(digit) for digit in ean8]
            ean13_digits = [int(digit) for digit in ean13]
            assert (sum(ean8_digits) + 2 * sum(ean8_digits[::2])) % 10 == 0
            assert (sum(ean13_digits) + 2 * sum(ean13_digits[1::2])) % 10 == 0

            self.assert_prefix(ean8_digits, provider.local_prefixes)
            self.assert_prefix(ean13_digits, provider.local_prefixes)

    def test_localized_ean8(self, faker, num_samples, provider):
        for _ in range(num_samples):
            ean8 = faker.localized_ean8()
            assert self.ean8_pattern.match(ean8)

            ean8_digits = [int(digit) for digit in ean8]
            assert (sum(ean8_digits) + 2 * sum(ean8_digits[::2])) % 10 == 0

            self.assert_prefix(ean8_digits, provider.local_prefixes)

    def test_localized_ean13(self, faker, num_samples, provider):
        for _ in range(num_samples):
            ean13 = faker.localized_ean13()
            assert self.ean13_pattern.match(ean13)

            ean13_digits = [int(digit) for digit in ean13]
            assert (sum(ean13_digits) + 2 * sum(ean13_digits[1::2])) % 10 == 0

            self.assert_prefix(ean13_digits, provider.local_prefixes)


class _LocaleNorthAmericaMixin(_LocaleCommonMixin):
    upc_a_pattern: Pattern = re.compile(r"\d{12}")
    upc_e_pattern: Pattern = re.compile(r"[01]\d{7}")

    def test_upc_a(self, faker, num_samples):
        for _ in range(num_samples):
            upc_a = faker.upc_a()
            assert self.upc_a_pattern.fullmatch(upc_a)

            # Included check digit must be correct
            upc_a_digits = [int(digit) for digit in upc_a]
            assert (sum(upc_a_digits) + 2 * sum(upc_a_digits[::2])) % 10 == 0

    def test_upc_ae_mode(self, faker, num_samples):
        for _ in range(num_samples):
            upc_ae = faker.upc_a(upc_ae_mode=True)
            assert self.upc_a_pattern.fullmatch(upc_ae)

            # Included check digit must be correct
            upc_ae_digits = [int(digit) for digit in upc_ae]
            assert (sum(upc_ae_digits) + 2 * sum(upc_ae_digits[::2])) % 10 == 0

    def test_upc_e_explicit_number_system(self, faker, num_samples):
        for _ in range(num_samples):
            upc_e_0 = faker.upc_e(number_system_digit=0)
            upc_e_1 = faker.upc_e(number_system_digit=1)
            assert self.upc_e_pattern.fullmatch(upc_e_0)
            assert self.upc_e_pattern.fullmatch(upc_e_1)
            assert upc_e_0[0] == "0"
            assert upc_e_1[0] == "1"

    def test_upc_e_safe_mode(self, faker):
        # For this test, we explicitly specify a base and a number system digit
        # so we do not have to wait for RNG to produce the right combinations.
        for _ in range(100):
            # Be aware that there are other unsafe combinations
            unsafe_base = f"{faker.random_int(0, 99):02}000{faker.random_int(3, 4)}"
            safe_base = unsafe_base[:2] + "0000"
            number_system_digit = faker.random_int(0, 1)

            # Safe mode will create a UPC-E barcode with the safe base
            # even if an unsafe base was supplied
            upc_e_safe = faker.upc_e(
                base=unsafe_base,
                number_system_digit=number_system_digit,
                safe_mode=True,
            )
            assert upc_e_safe[1:-1] == safe_base
            assert upc_e_safe[1:-1] != unsafe_base

            # Unsafe mode will force create a UPC-E barcode with unsafe base
            upc_e_unsafe = faker.upc_e(
                base=unsafe_base,
                number_system_digit=number_system_digit,
                safe_mode=False,
            )
            assert upc_e_unsafe[1:-1] != safe_base
            assert upc_e_unsafe[1:-1] == unsafe_base

            # What will be the same are their number system and check digits
            assert upc_e_safe[0] == upc_e_unsafe[0]
            assert upc_e_safe[-1] == upc_e_unsafe[-1]

    def test_upc_a2e_bad_values(self, faker, provider):
        # Invalid data type
        with pytest.raises(TypeError):
            provider._convert_upc_a2e(12345678)

        # Invalid string
        with pytest.raises(ValueError):
            provider._convert_upc_a2e("abcdef")

    def test_upc_a2e2a(self, faker, num_samples, provider):
        for _ in range(num_samples):
            upc_a = faker.upc_a(upc_ae_mode=True)
            assert self.upc_a_pattern.fullmatch(upc_a)

            # Convert UPC-A to UPC-E
            upc_e = provider._convert_upc_a2e(upc_a)

            # Number system and check digits must be the same
            assert int(upc_a[0]) == int(upc_e[0])
            assert int(upc_a[-1]) == int(upc_e[-1])

            # Create a new UPC-A barcode based on the UPC-E barcode
            new_upc_a = faker.upc_a(upc_ae_mode=True, base=upc_e[1:-1], number_system_digit=int(upc_e[0]))

            # New UPC-A barcode must be the same as the original
            assert upc_a == new_upc_a

    def test_upc_e2a2e(self, faker, num_samples, provider):
        for _ in range(num_samples):
            upc_e = faker.upc_e()
            assert self.upc_e_pattern.fullmatch(upc_e)

            # Create a new UPC-A barcode based on the UPC-E barcode
            upc_a = faker.upc_a(upc_ae_mode=True, base=upc_e[1:-1], number_system_digit=int(upc_e[0]))

            # Number system and check digits must be the same
            assert int(upc_a[0]) == int(upc_e[0])
            assert int(upc_a[-1]) == int(upc_e[-1])

            # Convert UPC-A to UPC-E
            new_upc_e = provider._convert_upc_a2e(upc_a)

            # New UPC-E barcode must be the same as the original
            assert new_upc_e == upc_e


class TestEnUs(_LocaleNorthAmericaMixin):
    """Tests en_US barcode provider"""

    num_samples = 1000

    @staticmethod
    def get_provider_class():
        from faker.providers.barcode.en_US import Provider

        return Provider


class TestEnCa(_LocaleNorthAmericaMixin):
    """Tests en_CA barcode provider"""

    num_samples = 1000

    @staticmethod
    def get_provider_class():
        from faker.providers.barcode.en_CA import Provider

        return Provider


class TestFrCa(_LocaleNorthAmericaMixin):
    """Tests fr_CA barcode provider"""

    num_samples = 1000

    @staticmethod
    def get_provider_class():
        from faker.providers.barcode.fr_CA import Provider

        return Provider


class TestJaJp(_LocaleCommonMixin):
    """Tests ja_JP barcode provider"""

    num_samples = 1000

    @staticmethod
    def get_provider_class():
        from faker.providers.barcode.ja_JP import Provider

        return Provider

    def test_jan(self, faker, num_samples, provider):
        for _ in range(num_samples):
            jan8 = faker.jan(8)
            jan13 = faker.jan(13)
            assert self.ean8_pattern.match(jan8)
            assert self.ean13_pattern.match(jan13)

            jan8_digits = [int(digit) for digit in jan8]
            jan13_digits = [int(digit) for digit in jan13]
            assert (sum(jan8_digits) + 2 * sum(jan8_digits[::2])) % 10 == 0
            assert (sum(jan13_digits) + 2 * sum(jan13_digits[1::2])) % 10 == 0

            self.assert_prefix(jan8_digits, provider.local_prefixes)
            self.assert_prefix(jan13_digits, provider.local_prefixes)

    def test_jan8(self, faker, num_samples, provider):
        for _ in range(num_samples):
            jan8 = faker.jan(8)
            assert self.ean8_pattern.match(jan8)

            jan8_digits = [int(digit) for digit in jan8]
            assert (sum(jan8_digits) + 2 * sum(jan8_digits[::2])) % 10 == 0

            self.assert_prefix(jan8_digits, provider.local_prefixes)

    def test_jan13(self, faker, num_samples, provider):
        for _ in range(num_samples):
            jan13 = faker.jan(13)
            assert self.ean13_pattern.match(jan13)

            jan13_digits = [int(digit) for digit in jan13]
            assert (sum(jan13_digits) + 2 * sum(jan13_digits[1::2])) % 10 == 0

            self.assert_prefix(jan13_digits, provider.local_prefixes)


class TestEsEs(_LocaleCommonMixin):
    """Tests es_ES barcode provider"""

    num_samples = 1000

    @staticmethod
    def get_provider_class():
        from faker.providers.barcode.es_ES import Provider

        return Provider

    def test_localized_ean(self, faker, num_samples, provider):
        for _ in range(num_samples):
            ean8 = faker.localized_ean(8)
            ean13 = faker.localized_ean(13)
            assert self.ean8_pattern.match(ean8)
            assert self.ean13_pattern.match(ean13)

            ean8_digits = [int(digit) for digit in ean8]
            ean13_digits = [int(digit) for digit in ean13]
            assert (sum(ean8_digits) + 2 * sum(ean8_digits[::2])) % 10 == 0
            assert (sum(ean13_digits) + 2 * sum(ean13_digits[1::2])) % 10 == 0

            self.assert_prefix(ean8_digits, provider.local_prefixes)
            self.assert_prefix(ean13_digits, provider.local_prefixes)



from ..ar_AA import Provider as ArabicPersonProvider


class Provider(ArabicPersonProvider):
    last_names = (
        "أبو اسنينة",
        "أبو شقدم",
        "أبو شلبك",
        "أبو غليون",
        "أبو قمر",
        "أستيتية",
        "الأدغم",
        "الإغباري",
        "البرغوثي",
        "التركمان",
        "التميمي",
        "الجنيدي",
        "الحسيني",
        "الحنبلي",
        "الخازن",
        "الخماش",
        "الخياط",
        "الزيتاوي",
        "الزيدانية",
        "السكاكيني",
        "الصالحي",
        "النشاشيبي",
        "النعنيش",
        "بدر",
        "ترابين",
        "جرار",
        "جزار",
        "حمامي",
        "حوسة",
        "خوري",
        "دغمش",
        "دلاشة",
        "شاهين",
        "صليبا",
        "طوقان",
        "فطاير",
        "قرادة",
        "كسواني",
        "مرمش",
        "مهيار",
        "نسيبة",
        "هاشم",
    )



from ..en_CA import Provider as BarcodeProvider


class Provider(BarcodeProvider):
    """Implement bank provider for ``fr_CA`` locale.

    There is no difference from the ``en_CA`` implementation.
    """

    pass



from typing import Optional

from .. import Provider as BankProvider


class Provider(BankProvider):
    """
    Implement bank provider for ``bn_BD`` locale.
    Sources:
        - https://wise.com/gb/swift-codes/BBHOBDDHXXX
        - https://www.banksbd.org/swift-codes.html
    """

    bban_format: str = "????#########"
    country_code = "BD"
    swift_location_codes = ("DH",)
    swift_branch_codes = (
        "ABBL",
        "AGBK",
        "ALAR",
        "ALFH",
        "BCBL",
        "BDDB",
        "BKBA",
        "BKSI",
        "BALB",
        "BRAK",
        "BBSH",
        "BSON",
        "CITI",
        "CCEY",
        "COYM",
        "CIBL",
        "DHBL",
        "DBBL",
        "EBLD",
        "EXBK",
        "FSEB",
        "FRMS",
        "HABB",
        "HSBC",
        "HVBK",
        "IFIC",
        "IBBL",
        "JAMU",
        "JANB",
        "MGBL",
        "MBLB",
        "MDBL",
        "MODH",
        "MTBL",
        "NGBL",
        "NBLB",
        "NBPA",
        "NCCL",
        "NRBD",
        "NRBB",
        "ONEB",
        "PRBL",
        "PRMR",
        "PUBA",
        "RUPB",
        "SJBL",
        "SOIV",
        "SBAC",
        "SEBD",
        "SDBL",
        "SCBL",
        "SBIN",
        "TTBL",
        "UBLD",
        "UCBL",
        "UTBL",
    )

    def swift8(self, use_dataset: bool = True) -> str:
        return super(self.__class__, self).swift8(use_dataset=use_dataset)

    def swift11(self, primary: bool = False, use_dataset: bool = True) -> str:
        return super(self.__class__, self).swift11(primary=primary, use_dataset=use_dataset)

    def swift(self, length: Optional[int] = None, primary: bool = False, use_dataset: bool = True) -> str:
        return super(self.__class__, self).swift(length=length, primary=primary, use_dataset=use_dataset)



import re
import string

from collections import OrderedDict

import pytest

from faker.providers import BaseProvider


class TestBaseProvider:
    """Test base provider methods"""

    def test_locale(self, faker, num_samples):
        locales = [
            f"{language}_{region}"
            for language, regions in BaseProvider.language_locale_codes.items()
            for region in regions
        ]
        for _ in range(num_samples):
            locale = faker.locale()
            assert locale in locales

    def test_language_code(self, faker, num_samples):
        language_codes = list(BaseProvider.language_locale_codes)
        for _ in range(num_samples):
            language_code = faker.language_code()
            assert language_code in language_codes

    def test_random_digit(self, faker, num_samples):
        samples = [faker.random_digit() for _ in range(num_samples * 10)]
        assert set(samples) == set(range(10))

    def test_random_digit_not_null(self, faker, num_samples):
        samples = [faker.random_digit_not_null() for _ in range(num_samples * 10)]
        assert set(samples) == set(range(1, 10))

    def test_random_digit_above_two(self, faker, num_samples):
        samples = [faker.random_digit_above_two() for _ in range(num_samples * 10)]
        assert set(samples) == set(range(2, 10))

    def test_random_digit_or_empty(self, faker, num_samples):
        expected = set(range(10))
        expected.add("")
        samples = [faker.random_digit_or_empty() for _ in range(num_samples * 10)]
        assert set(samples) == expected

    def test_random_digit_not_null_or_empty(self, faker, num_samples):
        expected = set(range(1, 10))
        expected.add("")
        samples = [faker.random_digit_not_null_or_empty() for _ in range(num_samples * 10)]
        assert set(samples) == expected

    def test_random_number(self, faker):
        number = faker.random_number(10, True)
        assert len(str(number)) == 10

        # Digits parameter < 0
        with pytest.raises(ValueError):
            number = faker.random_number(-1, True)

        # Digits parameter < 1 with fix_len=True
        with pytest.raises(ValueError):
            number = faker.random_number(0, True)

    @pytest.mark.parametrize(
        "text,pattern",
        [
            ("", r""),
            ("abcd", r"abcd"),
            ("#" * 100, r"[0-9]{100}"),
            ("%" * 100, r"[1-9]{100}"),
            ("!" * 100, r"[0-9]{,100}"),
            ("@" * 100, r"[0-9]{,100}"),
            (
                "##!abc %%@def##!" * 100,
                r"(?:[0-9]{2,3}abc [1-9]{2,3}def[0-9]{2,3}){100}",
            ),
            (
                "#@@#^?あ5漢!!%%@" * 100,
                r"(?:\d[1-9]{,2}\d\^\?あ5漢\d{,2}[1-9]{2}[1-9]*){100}",
            ),
        ],
        ids=[
            "empty_string",
            "no_valid_placeholders",
            "only_number_signs",
            "only_percent_signs",
            "only_exclamation_marks",
            "only_at_symbols",
            "with_ascii_characters",
            "with_other_symbols_and_non_ascii",
        ],
    )
    def test_numerify(self, faker, num_samples, text, pattern):
        for _ in range(num_samples):
            numerified = faker.numerify(text)
            assert re.fullmatch(pattern, numerified)

    @pytest.mark.parametrize(
        "text,letters,pattern",
        [
            ("", string.ascii_letters, r""),
            ("abcd", string.ascii_letters, r"abcd"),
            ("???", string.ascii_letters, r"[0-9a-zA-Z]{3}"),
            ("???", "aBcDeFgHiJ12345", r"[1-5aBcDeFgHiJ]{3}"),
            ("??Xr^#7p??", "AbCdخあ5漢7Я", r"[AbCdخあ5漢7Я]{2}Xr\^#7p[AbCdخあ5漢7Я]{2}"),
        ],
        ids=[
            "empty_string",
            "no_valid_placeholders",
            "letters_using_whole_ascii",
            "letters_using_ascii_subset",
            "pattern_with_other_symbols_and_letters_using_non_ascii",
        ],
    )
    def test_lexify(self, faker, num_samples, text, letters, pattern):
        for _ in range(num_samples):
            lexified = faker.lexify(text, letters=letters)
            assert re.fullmatch(pattern, lexified)

    @pytest.mark.parametrize(
        "text,letters,pattern",
        [
            ("", string.ascii_letters, r""),
            ("abcd", string.ascii_letters, r"abcd"),
            ("???", string.ascii_letters, r"[0-9a-zA-Z]{3}"),
            ("???", "aBcDeFgHiJ12345", r"[1-5aBcDeFgHiJ]{3}"),
            ("#%!@???", string.ascii_letters, r"\d[1-9]\d*[1-9]*[0-9a-zA-Z]{3}"),
            ("#%!@???", "aBcDeFgHiJ12345", r"\d[1-9]\d*[1-9]*[1-5aBcDeFgHiJ]{3}"),
            (
                "#%!@??Xr7p??",
                "AbCdخあ5漢7Я",
                r"\d[1-9]\d*[1-9]*[AbCdخあ5漢7Я]{2}Xr7p[AbCdخあ5漢7Я]{2}",
            ),
        ],
        ids=[
            "empty_string",
            "no_valid_placeholders",
            "simple_pattern_and_letters_using_whole_ascii",
            "simple_pattern_and_letters_using_ascii_subset",
            "more_complex_pattern_and_letters_using_whole_ascii",
            "more_complex_pattern_and_letters_using_ascii_subset",
            "more_complex_pattern_with_other_symbols_and_letters_using_non_ascii",
        ],
    )
    def test_bothify(self, faker, num_samples, text, letters, pattern):
        for _ in range(num_samples):
            bothified = faker.bothify(text, letters=letters)
            assert re.fullmatch(pattern, bothified)

    @pytest.mark.parametrize(
        "text,upper,pattern",
        [
            ("", False, r""),
            ("", True, r""),
            ("abcd", False, r"abcd"),
            ("abcd", True, r"abcd"),
            ("^^^^", False, r"[0-9a-f]{4}"),
            ("^^^^", True, r"[0-9A-F]{4}"),
            (
                "Abc ^^^ %^^^?あ5漢!#^^",
                False,
                r"Abc [0-9a-f]{3} %[0-9a-f]{3}\?あ5漢!#[0-9a-f]{2}",
            ),
            (
                "Abc ^^^ %^^^?あ5漢!#^^",
                True,
                r"Abc [0-9A-F]{3} %[0-9A-F]{3}\?あ5漢!#[0-9A-F]{2}",
            ),
        ],
        ids=[
            "empty_string_lowercase",
            "empty_string_uppercase",
            "no_circumflex_lowercase",
            "no_circumflex_uppercase",
            "simple_pattern_lowercase",
            "simple_pattern_uppercase",
            "complex_pattern_lowercase",
            "complex_pattern_uppercase",
        ],
    )
    def test_hexify(self, faker, num_samples, text, upper, pattern):
        for _ in range(num_samples):
            hexified = faker.hexify(text, upper=upper)
            assert re.fullmatch(pattern, hexified)

    def test_random_letter(self, faker, num_samples):
        for _ in range(num_samples):
            letter = faker.random_letter()
            assert letter.isalpha()

    @pytest.mark.parametrize(
        "length",
        [0, 1, 2],
        ids=[
            "empty_list",
            "list_with_one_element",
            "list_with_two_elements",
        ],
    )
    def test_random_letters(self, faker, length):
        letters = faker.random_letters(length=length)
        assert len(letters) == length
        assert isinstance(letters, list)
        for letter in letters:
            assert letter.isalpha()

    def test_random_lowercase_letter(self, faker, num_samples):
        for _ in range(num_samples):
            letter = faker.random_lowercase_letter()
            assert letter.isalpha() and letter.lower() == letter

    def test_random_uppercase_letter(self, faker, num_samples):
        for _ in range(num_samples):
            letter = faker.random_uppercase_letter()
            assert letter.isalpha() and letter.upper() == letter

    def test_random_element(self, faker, num_samples):
        # dicts not allowed because they introduce dependency on PYTHONHASHSEED
        with pytest.raises(ValueError):
            faker.random_element({})

        choices = ("a", "b", "c", "d")
        for _ in range(num_samples):
            assert faker.random_element(choices) in choices

        choices = OrderedDict([("a", 5), ("b", 2), ("c", 2), ("d", 1)])
        for _ in range(num_samples):
            assert faker.random_element(choices) in choices

        choices = OrderedDict([("a", 0.5), ("b", 0.2), ("c", 0.2), ("d", 0.1)])
        for _ in range(num_samples):
            assert faker.random_element(choices) in choices

    def test_random_sample(self, faker):
        # Too many items requested
        with pytest.raises(ValueError):
            faker.random_sample("abcde", 6)

        # Same length
        sample = faker.random_sample("abcd", 4)
        assert sorted(sample) == list("abcd")

        sample = faker.random_sample("abcde", 5)
        assert sorted(sample) == list("abcde")

        # Length = 3
        sample = faker.random_sample("abcde", 3)
        assert len(sample) == 3
        assert set(sample).issubset(set("abcde"))

        # Length = 1
        sample = faker.random_sample("abcde", 1)
        assert len(sample) == 1
        assert set(sample).issubset(set("abcde"))

        # Length = 0
        sample = faker.random_sample("abcde", 0)
        assert sample == []

    def test_randomize_nb_elements(self, faker, num_samples):
        assert faker.randomize_nb_elements(number=1, le=True, ge=True) == 1
        assert faker.randomize_nb_elements(le=True, ge=True) == 10
        assert faker.randomize_nb_elements(min=42) == 42
        assert faker.randomize_nb_elements(max=1) == 1

        number = 9999
        lower_bound = int(number * 0.6)
        upper_bound = int(number * 1.4)

        for _ in range(num_samples):
            res = faker.randomize_nb_elements(number=number, le=True)
            assert res >= lower_bound
            assert res <= number, f"{res!r} is not <= than {number!r}"

        for _ in range(num_samples):
            res = faker.randomize_nb_elements(number=number, ge=True)
            assert number <= res <= upper_bound

        for _ in range(num_samples):
            res = faker.randomize_nb_elements(number=number)
            assert lower_bound <= res <= upper_bound



from pathlib import Path

import pytest

EXCLUSIVE_SESSION_FLAG = "--exclusive-faker-session"
SKIP_REASON = (
    f"This test is skipped by default since it depends on changes in the behavior of session-scoped fixtures. "
    f'Use a separate pytest run for tests like this with the "{EXCLUSIVE_SESSION_FLAG}" flag specified.'
)


def pytest_addoption(parser):
    parser.addoption(
        EXCLUSIVE_SESSION_FLAG,
        action="store_true",
        default=False,
        help="Allows tests that require an exclusive session to run",
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption(EXCLUSIVE_SESSION_FLAG):
        return
    skip_lacks_exclusive_session = pytest.mark.skip(reason=SKIP_REASON)
    session_overrides_dir = Path(__file__).resolve().parent
    for item in items:
        if str(item.fspath).startswith(str(session_overrides_dir)):
            item.add_marker(skip_lacks_exclusive_session)
